import pytest
import numpy as np
from rbc.tests import heavydb_fixture, _TestTable


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(6, 4),
                             suffices=['linestring', 'multilinestring',
                                       'multipoint',
                                       'polygon', 'multipolygon']):
        define(o)
        yield o


def define(heavydb):
    @heavydb("int32(TableFunctionManager, Column<Z>, OutputColumn<Z>)",
             Z=['GeoLineString', 'GeoMultiLineString', 'GeoMultiPoint',
                'GeoPolygon', 'GeoMultiPolygon'],
             devices=['cpu'])
    def rbc_ct_copy(mgr, inputs, outputs):
        size = len(inputs)
        mgr.set_output_item_values_total_number(0, inputs.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if inputs.is_null(i):
                outputs.set_null(i)
            else:
                outputs[i] = inputs[i]
        return size

    @heavydb("int32(TableFunctionManager, Column<T>, int64, OutputColumn<Z>, OutputColumn<Z>)",
             T=['GeoMultiPoint', 'GeoLineString'], Z=['double'], devices=['cpu'])
    def rbc_ct_pointn(mgr, points, n, xcoords, ycoords):
        size = len(points)
        mgr.set_output_row_size(size)
        for i in range(size):
            if points.is_null(i):
                xcoords.set_null(i)
                ycoords.set_null(i)
            else:
                point = points[i][n - 1]  # n is one-based
                xcoords[i] = point.x
                ycoords[i] = point.y
        return size

    @heavydb("int32(TableFunctionManager, Cursor<T, T, T>, OutputColumn<Z>, OutputColumn<int>)",
             T=['GeoLineString'], Z=['GeoPolygon'], devices=['cpu'])
    def rbc_ct_make_polygon3(mgr, rings, holes1, holes2, polygons, sizes):
        size = len(rings)
        mgr.set_output_item_values_total_number(0,
                                                rings.get_n_of_values() +
                                                holes1.get_n_of_values() +
                                                holes2.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if rings.is_null(i):
                polygons.set_null(i)
                sizes.set_null(i)
            else:
                polygon_coords = []
                polygon_coords.append(rings[i].to_coords())
                if not holes1.is_null(i):
                    polygon_coords.append(holes1[i].to_coords())
                    if not holes2.is_null(i):
                        polygon_coords.append(holes2[i].to_coords())

                polygon = polygons[i]
                polygon.from_coords(polygon_coords)

                n_of_points = 0
                for j in range(len(polygon)):
                    n_of_points += len(polygon[j])
                sizes[i] = n_of_points
        return size


class _GeoPolygonTest(_TestTable):
    @property
    def suffix(self):
        return 'geo_polygon_test'

    @property
    def sqltypes(self):
        return ('LINESTRING',
                'LINESTRING',
                'LINESTRING',
                'GEOMETRY(LINESTRING, 4326)',
                'GEOMETRY(LINESTRING, 4326)',
                'GEOMETRY(LINESTRING, 4326)',
                'GEOMETRY(LINESTRING, 4326) ENCODING NONE',
                'GEOMETRY(LINESTRING, 4326) ENCODING NONE',
                'GEOMETRY(LINESTRING, 4326) ENCODING NONE',
                'GEOMETRY(LINESTRING, 900913)',
                'GEOMETRY(LINESTRING, 900913)',
                'GEOMETRY(LINESTRING, 900913)')

    @property
    def values(self):
        return {
            'r1': ['LINESTRING(1 2,3 4,5 6,7 8,9 10)', 'LINESTRING(0 0,5 0,5 5,0 5)',
                   'LINESTRING(1 2,3 4,5 6,7 8,9 10)', None],
            'h1': ['LINESTRING(2 3,3 4,1 2)', None, 'LINESTRING(2 3,3 4,1 2)', None],
            'hh1': [None, None, 'LINESTRING(9 10,5 6,7 8)', None],
            'r2': ['LINESTRING(0 0,5 0,5 5,0 5)', 'LINESTRING(0 0,7 0,7 7,0 7)',
                   'LINESTRING(0 0,5 0,5 5,0 5)', None],
            'h2': ['LINESTRING(2 2,2 1,1 1,1 2)', None,
                   'LINESTRING(2 2,2 1,1 1,1 2)', None],
            'hh2': [None, None, 'LINESTRING(0 0,0 1,1 0)', None],
            'r3': ['LINESTRING(0 0,6 0,6 6,0 6))', 'LINESTRING(0 0,7 0,7 7,0 7)',
                   'LINESTRING(0 0,6 0,6 6,0 6))', None],
            'h3': ['LINESTRING(3 3,3 2,2 2,2 3)', None,
                   'LINESTRING(3 3,3 2,2 2,2 3)', None],
            'hh3': [None, None, 'LINESTRING(0 0,0 1,1 0)', None],
            'r4': ['LINESTRING(0 0,7 0,7 7,0 7)', 'LINESTRING(0 0,4 0,4 4,0 4)',
                   'LINESTRING(0 0,7 0,7 7,0 7)', None],
            'h4': ['LINESTRING(4 4,4 2,2 3,2 4)', None,
                   'LINESTRING(4 4,4 2,2 3,2 4)', None],
            'hh4': [None, None, 'LINESTRING(0 0,0 1,1 0)', None],
        }


@pytest.fixture(scope='module')
def create_table(heavydb):
    table_name = 'geo_polygon_test'
    geo = _GeoPolygonTest()

    heavydb.sql_execute(f'DROP TABLE IF EXISTS {table_name}')
    heavydb.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name} ({geo.table_defn})')
    heavydb.load_table_columnar(f'{table_name}', **geo.values)
    yield heavydb
    heavydb.sql_execute(f'DROP TABLE IF EXISTS {table_name}')


@pytest.mark.parametrize('suffix,col', [
    ('linestring', 'l1'),
    ('linestring', 'l2'),
    ('linestring', 'l3'),
    ('linestring', 'l4'),
    ('multilinestring', 'ml1'),
    ('multilinestring', 'ml2'),
    ('multilinestring', 'ml3'),
    ('multilinestring', 'ml4'),
    ('multipoint', 'mp1'),
    ('multipoint', 'mp2'),
    ('multipoint', 'mp3'),
    ('multipoint', 'mp4'),
    ('polygon', 'p1'),
    ('polygon', 'p2'),
    ('polygon', 'p3'),
    ('polygon', 'p4'),
    ('multipolygon', 'mp1'),
    ('multipolygon', 'mp2'),
    ('multipolygon', 'mp3'),
    ('multipolygon', 'mp4'),
])
def test_ct_copy(heavydb, suffix, col):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    query = (f'select * from table(rbc_ct_copy(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_copy(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('n', ('1',))
@pytest.mark.parametrize('suffix,col', [
    ('linestring', 'l1'),
    ('linestring', 'l2'),
    ('linestring', 'l3'),
    ('linestring', 'l4'),
    ('multipoint', 'mp1'),
    pytest.param('multipoint', 'mp2', marks=pytest.mark.xfail(reason='odd bug')),
    ('multipoint', 'mp3'),
    ('multipoint', 'mp4'),
])
def test_ct_pointn(heavydb, suffix, col, n):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    query = (f'select * from table(rbc_ct_pointn(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), {n}));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_pointn(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), {n}));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.usefixtures('create_table')
@pytest.mark.parametrize('rcol, hcol, hhcol', [
    ('r1', 'h1', 'hh1'),
    ('r2', 'h2', 'hh2'),
    ('r3', 'h3', 'hh3'),
    ('r4', 'h4', 'hh4'),
])
def test_ct_make_polygon3(heavydb, rcol, hcol, hhcol):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    table = 'geo_polygon_test'
    func = 'ct_make_polygon3'

    query = (f'select * from table(rbc_{func}(cursor(select {rcol}, {hcol}, {hhcol} '
             f'from {table})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table({func}(cursor(select {rcol}, {hcol}, {hhcol} '
             f'from {table})));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)



from rbc.tests import heavydb_fixture
from rbc.heavydb import TextEncodingNone
import pytest

rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), suffices=['text']):
        yield o


def test_table_data(heavydb):
    heavydb.reset()

    descr, result = heavydb.sql_execute(
        f'select t4, t2, t1, s, n from {heavydb.table_name}text')
    result = list(result)

    assert result == [('foofoo', 'foofoo', 'fun', ['foo', 'bar'], 'fun'),
                      ('bar', 'bar', 'bar', ['fun', 'bar'], 'bar'),
                      ('fun', 'fun', 'foo', ['foo'], 'foo'),
                      ('bar', 'bar', 'barr', ['foo', 'bar'], 'barr'),
                      ('foo', 'foo', 'foooo', ['fun', 'bar'], 'foooo')]


def test_TextEncodingNone_len(heavydb):
    heavydb.reset()

    @heavydb('int32(TextEncodingNone, TextEncodingNone)')
    def mystrlen(s, s2):
        return len(s) * 100 + len(s2)

    sql_query_expected = f"select length(n) * 100 + length(n2) from {heavydb.table_name}text"
    sql_query = f"select mystrlen(n, n2) from {heavydb.table_name}text"

    descr, result_expected = heavydb.sql_execute(sql_query_expected)
    result_expected = list(result_expected)
    descr, result = heavydb.sql_execute(sql_query)
    result = list(result)

    assert result == result_expected


def test_TextEncodingNone_ord(heavydb):
    heavydb.reset()

    @heavydb('int64(TextEncodingNone, int32)', devices=['gpu', 'cpu'])
    def myord(s, i):
        return s[i] if i < len(s) else 0

    sql_query_data = f"select n from {heavydb.table_name}text"
    descr, data = heavydb.sql_execute(sql_query_data)
    data = list(data)

    for i in range(5):
        result_expected = [(ord(d[0][i]) if i < len(d[0]) else 0, ) for d in data]
        sql_query = f"select myord(n, {i}) from {heavydb.table_name}text"
        descr, result = heavydb.sql_execute(sql_query)
        result = list(result)
        assert result == result_expected


def test_TextEncodingNone_return(heavydb):
    heavydb.reset()

    @heavydb('TextEncodingNone(int32, int32)')
    def make_abc(first, n):
        r = TextEncodingNone(n)
        for i in range(n):
            r[i] = first + i
        return r

    sql_query = "select make_abc(97, 10)"
    descr, result = heavydb.sql_execute(sql_query)
    result = list(result)

    assert result == [('abcdefghij', )]


def test_TextEncodingNone_upper(heavydb):
    heavydb.reset()

    @heavydb('TextEncodingNone(TextEncodingNone)')
    def myupper(s):
        r = TextEncodingNone(len(s))
        for i in range(len(s)):
            c = s[i]
            if c >= 97 and c <= 122:
                c = c - 32
            r[i] = c
        return r

    sql_query = f"select n, myupper(n) from {heavydb.table_name}text"
    descr, result = heavydb.sql_execute(sql_query)
    result = list(result)

    for n, u in result:
        assert n.upper() == u


def test_TextEncodingNone_str_constructor(heavydb):
    heavydb.reset()

    @heavydb('TextEncodingNone(int32)')
    def constructor(_):
        return TextEncodingNone('hello world')

    assert constructor(3).execute() == 'hello world'


def test_TextEncodingNone_eq(heavydb):
    heavydb.require_version((5, 8), 'Requires heavydb 5.8 or newer')

    heavydb.reset()

    @heavydb('int32(TextEncodingNone, TextEncodingNone)')
    def eq1(a, b):
        return a == b

    @heavydb('int32(TextEncodingNone)', devices=['CPU'])
    def eq2(a):
        return a == 'world'

    assert eq1('hello', 'hello').execute() == 1
    assert eq1('c', 'c').execute() == 1
    assert eq1('hello', 'h').execute() == 0
    assert eq1('hello', 'hello2').execute() == 0

    assert eq2('world').execute() == 1


def test_TextEncodingNone_ne(heavydb):
    heavydb.require_version((5, 8), 'Requires heavydb 5.8 or newer')

    heavydb.reset()

    @heavydb('int32(TextEncodingNone, TextEncodingNone)')
    def ne1(a, b):
        return a != b

    @heavydb('int32(TextEncodingNone)', devices=['CPU'])
    def ne2(a):
        return a != 'world'

    assert ne1('hello', 'hello').execute() == 0
    assert ne1('c', 'c').execute() == 0
    assert ne1('hello', 'h').execute() == 1
    assert ne1('hello', 'hello2').execute() == 1

    assert ne2('world').execute() == 0


def test_if_else_assignment(heavydb):
    heavydb.require_version((5, 8), 'Requires heavydb 5.8 or newer')

    # fix for issue #487

    @heavydb("TextEncodingNone(double)")
    def classify_slope(slope):
        if slope <= 5:
            res = TextEncodingNone("low")
        elif 5 < slope < 15:
            res = TextEncodingNone("med")
        else:
            res = TextEncodingNone("high")
        return res

    assert classify_slope(2.4).execute() == "low"
    assert classify_slope(5.4).execute() == "med"
    assert classify_slope(15.4).execute() == "high"


def test_return_text(heavydb):
    @heavydb("TextEncodingNone(TextEncodingNone)", devices=['cpu'])
    def fn(t):
        return t

    table = heavydb.table_name + 'text'
    _, result = heavydb.sql_execute(f"select n, fn(n) from {table};")
    result = list(zip(*result))
    expected, got = result
    assert expected == got


def test_to_string(heavydb):

    @heavydb("TextEncodingNone(TextEncodingNone)", devices=['cpu'])
    def to_string(t):
        s = t.to_string()
        return TextEncodingNone(s)

    table = heavydb.table_name + 'text'
    _, result = heavydb.sql_execute(f"select n, to_string(n) from {table};")
    result = list(zip(*result))
    expected, got = result
    assert expected == got



__all__ = ['allocate_varlen_buffer']

from numba.core import cgutils
from llvmlite import ir


def allocate_varlen_buffer(builder, element_count, element_size):
    """
    Allocates ``(element_count + 1) * element_size`` bytes

    When an LLVM module defines manage_memory_buffer=1 (default), the memory
    buffers allocated by "allocate_varlen_buffer" will be freed at the
    execution clean-up stage.
    """
    i8p = ir.IntType(8).as_pointer()
    i64 = ir.IntType(64)

    module = builder.module
    name = 'allocate_varlen_buffer'
    fnty = ir.FunctionType(i8p, [i64, i64])
    fn = cgutils.get_or_insert_function(module, fnty, name)
    return builder.call(fn, [element_count, element_size])



import pytest
import numpy as np
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(6, 4),
                             suffices=['multipolygon']):
        define(o)
        yield o


def define(heavydb):
    @heavydb("int32(TableFunctionManager, Column<Z>, int64, OutputColumn<K>)",
             Z=['GeoMultiPolygon'], K=['GeoPolygon'], devices=['cpu'])
    def rbc_ct_polygonn(mgr, mpolygons, n, polygons):
        size = len(mpolygons)
        mgr.set_output_item_values_total_number(0, mpolygons.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if mpolygons.is_null(i):
                polygons.set_null(i)
            else:
                polygons.set_item(i, mpolygons[i][n - 1])
        return size


@pytest.mark.parametrize('n', ('1',))
@pytest.mark.parametrize('suffix,col', [
    ('multipolygon', 'mp1'),
    ('multipolygon', 'mp2'),
    ('multipolygon', 'mp3'),
    ('multipolygon', 'mp4'),
])
def test_ct_polygonn(heavydb, suffix, col, n):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    query = (f'select * from table(rbc_ct_polygonn(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), {n}));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_polygonn(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), {n}));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)



"""
Array API specification for sorting functions.

https://data-apis.org/array-api/latest/API_specification/sorting_functions.html
"""
from rbc.stdlib import Expose

__all__ = ["argsort", "sort"]

expose = Expose(globals(), "sorting_functions")


@expose.not_implemented("argsort")
def _array_api_argsort(x, *, axis=-1, descending=False, stable=True):
    """
    Returns the indices that sort an array `x` along a specified axis.
    """
    pass


@expose.not_implemented("sort")
def _array_api_sort(x, *, axis=-1, descending=False, stable=True):
    """
    Returns a sorted copy of an input array `x`.
    """
    pass



import pytest
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(6, 3),
                             suffices=['array', 'arraynull', 'text']):
        define(o)
        yield o


def define(heavydb):
    @heavydb("int32(TableFunctionManager, Column<Array<T>> inp, OutputColumn<Array<T>> out | input_id=args<0>)",  # noqa: E501
             T=['float', 'int64', 'TextEncodingDict'], devices=['cpu'])
    def rbc_array_copy(mgr, inp, out):
        sz = len(inp)
        output_values_size = 0
        for i in range(sz):
            output_values_size += len(inp[i])

        # initialize array buffers
        mgr.set_output_array_values_total_number(0, output_values_size)

        mgr.set_output_row_size(sz)
        for i in range(sz):
            if inp.is_null(i):
                out.set_null(i)
            else:
                out.set_item(i, inp[i])
        return sz

    @heavydb('int32(TableFunctionManager, ColumnList<Array<T>> lst, OutputColumn<Array<T>> out | input_id=args<0>)',  # noqa: E501
             T=['int64', 'float', 'TextEncodingDict'], devices=['cpu'])
    def rbc_array_concat(mgr, lst, out):
        output_values_size = 0

        for j in range(lst.ncols):
            for i in range(lst.nrows):
                output_values_size += len(lst[j][i])

        mgr.set_output_array_values_total_number(
            0,  # output column index
            output_values_size,  # upper bound to the number of items
                                 # in all output arrays
        )

        size = lst.nrows
        mgr.set_output_row_size(size)

        for i in range(lst.nrows):
            for j in range(lst.ncols):
                col = lst[j]
                arr = col[i]
                out.concat_item(i, arr)

        return size


@pytest.mark.parametrize("suffix", ['array'])
@pytest.mark.parametrize("col", ['i8', 'f4'])
def test_copy(heavydb, suffix, col):
    if heavydb.version[:2] < (6, 3):
        pytest.skip('Requires HeavyDB 6.3 or newer')

    query = (f'select * from table(rbc_array_copy(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select {col} from {heavydb.table_name}{suffix}')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    assert len(result) == len(expected) == 5
    assert result == expected


@pytest.mark.parametrize("col", ['s', 'na'])
def test_copy_text(heavydb, col):
    if heavydb.version[:2] < (6, 3):
        pytest.skip('Requires HeavyDB 6.3 or newer')

    if col == 'na':
        pytest.skip('Column<Array<TextEncodingNone>> is not supported yet.')

    query = (f'select * from table(rbc_array_copy(cursor(select {col} '
             f'from {heavydb.table_name}text)));')
    _, result = heavydb.sql_execute(query)

    query = (f'select {col} from {heavydb.table_name}text')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    assert len(result) == len(expected) == 5
    assert result == expected


@pytest.mark.parametrize("suffix", ['array', 'arraynull'])
@pytest.mark.parametrize("col", ['i8', 'f4'])
def test_concat(heavydb, suffix, col):
    if heavydb.version[:2] < (6, 3):
        pytest.skip('Requires HeavyDB 6.3 or newer')

    query = (f'select * from table(rbc_array_concat(cursor(select {col}, {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select {col}, {col} from {heavydb.table_name}{suffix}')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    assert len(result) == len(expected) == 5
    for a, (b, c) in zip(result, expected):
        if a == (None,):
            assert b is None and c is None
        else:
            assert a == (b + c,)


@pytest.mark.parametrize("col", ['s', 'na'])
def test_concat_text(heavydb, col):
    if heavydb.version[:2] < (6, 3):
        pytest.skip('Requires HeavyDB 6.3 or newer')

    if col == 'na':
        pytest.skip('Column<Array<TextEncodingNone>> is not supported yet.')

    query = (f'select * from table(rbc_array_concat(cursor(select {col} '
             f'from {heavydb.table_name}text)));')
    _, result = heavydb.sql_execute(query)

    query = (f'select {col} from {heavydb.table_name}text')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    assert len(result) == len(expected) == 5
    assert result == expected



import os
from numba.core import cgutils
from numba.cpython.hashing import _hashsecret as hashsecret
from llvmlite import ir
from contextlib import contextmanager
from rbc.targetinfo import TargetInfo

void = ir.VoidType()
i8 = ir.IntType(8)
i8p = i8.as_pointer()
i8pp = i8p.as_pointer()
i32 = ir.IntType(32)
i64 = ir.IntType(64)

REFCT_IDX = 0
DTOR_IDX = 1
DTOR_INFO_IDX = 2
DATA_IDX = 3
SIZE_IDX = 4
EXTERNAL_ALLOCATOR_IDX = 5


# Function Names
NRT_MemInfo_init = "NRT_MemInfo_init"

NRT_MemInfo_alloc = "NRT_MemInfo_alloc"
NRT_MemInfo_alloc_safe = "NRT_MemInfo_alloc_safe"

NRT_MemInfo_alloc_dtor = "NRT_MemInfo_alloc_dtor"
NRT_MemInfo_alloc_dtor_safe = "NRT_MemInfo_alloc_dtor_safe"

NRT_Allocate = "NRT_Allocate"
NRT_Allocate_External = "NRT_Allocate_External"
NRT_MemInfo_call_dtor = "NRT_MemInfo_call_dtor"

NRT_incref = "NRT_incref"
NRT_decref = "NRT_decref"

NRT_MemInfo_data_fast = "NRT_MemInfo_data_fast"

NRT_MemInfo_new = "NRT_MemInfo_new"
NRT_MemInfo_new_varsize_dtor = "NRT_MemInfo_new_varsize_dtor"
NRT_MemInfo_varsize_alloc = "NRT_MemInfo_varsize_alloc"
NRT_MemInfo_varsize_realloc = "NRT_MemInfo_varsize_realloc"
NRT_MemInfo_varsize_free = "NRT_MemInfo_varsize_free"
NRT_MemInfo_new_varsize = "NRT_MemInfo_new_varsize"
NRT_MemInfo_alloc_safe_aligned = "NRT_MemInfo_alloc_safe_aligned"
NRT_Reallocate = "NRT_Reallocate"
NRT_Reallocate_NoCopy = "NRT_Reallocate_NoCopy"
NRT_dealloc = "NRT_dealloc"
NRT_Free = "NRT_Free"
NRT_MemInfo_destroy = "NRT_MemInfo_destroy"

nrt_internal_custom_dtor = "nrt_internal_custom_dtor"
nrt_allocate_meminfo_and_data = "nrt_allocate_meminfo_and_data"
allocate_varlen_buffer = "allocate_varlen_buffer"
malloc = "malloc"
free = "free"
realloc = "realloc"

nrt_debug_incr = "nrt_debug_incr"
nrt_debug_decr = "nrt_debug_decr"
nrt_debug_spaces = "nrt_debug_spaces"
nrt_global_var = "__nrt_global_var"


_debug_functions = (nrt_debug_incr, nrt_debug_decr, nrt_debug_spaces)


NULL = cgutils.get_null_value(cgutils.voidptr_t)


class RBC_NRT:
    def __init__(self, target_context, verbose=False):
        self.module = ir.Module(name="RBC_nrt")
        self.target_context = target_context
        self.verbose = verbose
        self.debug_nrt = int(os.environ.get("RBC_DEBUG_NRT", False))

        gv = ir.GlobalVariable(self.module, i64, nrt_global_var)
        gv.initializer = i64(0)

        self._define_types()
        self._define_functions()

        self.define()
        self.set_hashsecrets()

    def _define_types(self):
        ti = TargetInfo()

        # type_sizeof stores each size in bytes
        size_t = ir.IntType(ti.type_sizeof["size_t"] * 8)

        # As a reference, here's a copy of the MemInfo struct defined in
        # numba/core/context/nrt.cpp
        #
        # struct MemInfo {
        #     std::atomic_size_t     refct;
        #     NRT_dtor_function dtor;
        #     void              *dtor_info;
        #     void              *data;
        #     size_t            size;   /* only used for NRT allocated memory */
        #     void* external_allocator;
        # };
        #
        # And the LLVM representation of the struct above:
        #   { %"struct.std::atomic", void (i8*, i64, i8*)*, i8*, i8*, i64, i8* }
        #   %"struct.std::atomic" = type { %"struct.std::__atomic_base" }
        #   %"struct.std::__atomic_base" = type { i64 }

        MemInfo_t = ir.global_context.get_identified_type("Struct.MemInfo")
        # First argument must be "atomic_i64_t" instead of size_t
        if MemInfo_t.elements is None:
            MemInfo_t.set_body(size_t, i8p, i8p, i8p, size_t, i8p)
        MemInfo_ptr_t = MemInfo_t.as_pointer()

        self.size_t = size_t
        self.MemInfo_t = MemInfo_t
        self.MemInfo_ptr_t = MemInfo_ptr_t
        self.sizeof_MemInfo_t = i64(self.target_context.get_abi_sizeof(MemInfo_t))

    def _define_functions(self):
        size_t = self.size_t
        MemInfo_ptr_t = self.MemInfo_ptr_t

        self.NRT_DTOR_FUNCTION = ir.FunctionType(void, [i8p, size_t, i8p])

        self.function_types = {
            # defined in nrt.h
            NRT_MemInfo_init: (
                ir.FunctionType(void, [MemInfo_ptr_t, i8p, size_t, i8p, i8p, i8p]),
                ("mi", "data", "size", "dtor", "dtor_info", "external_allocator"),
            ),
            NRT_MemInfo_alloc: (ir.FunctionType(MemInfo_ptr_t, [size_t]), ("size",)),
            NRT_MemInfo_alloc_safe: (
                ir.FunctionType(MemInfo_ptr_t, [size_t]),
                ("size",),
            ),
            NRT_MemInfo_alloc_dtor: (
                ir.FunctionType(MemInfo_ptr_t, [size_t, i8p]),
                ("size", "dtor"),
            ),
            NRT_MemInfo_alloc_dtor_safe: (
                ir.FunctionType(MemInfo_ptr_t, [size_t, i8p]),
                ("size", "dtor"),
            ),
            NRT_Allocate: (ir.FunctionType(i8p, [size_t]), ("size",)),
            NRT_Allocate_External: (
                ir.FunctionType(i8p, [size_t, i8p]),
                ("size", "allocator"),
            ),
            NRT_MemInfo_call_dtor: (ir.FunctionType(void, [MemInfo_ptr_t]), ("mi",)),
            NRT_incref: (ir.FunctionType(void, [i8p]), ("ptr",)),
            NRT_decref: (ir.FunctionType(void, [i8p]), ("ptr",)),
            NRT_MemInfo_data_fast: (ir.FunctionType(i8p, [i8p]), ("meminfo_ptr",)),
            NRT_MemInfo_new: (
                ir.FunctionType(MemInfo_ptr_t, [i8p, size_t, i8p, i8p]),
                ("data", "size", "dtor", "dtor_info"),
            ),
            NRT_MemInfo_alloc_safe_aligned: (
                ir.FunctionType(MemInfo_ptr_t, [size_t, i32]),
                ("size", "aligned"),
            ),
            NRT_MemInfo_new_varsize: (
                ir.FunctionType(MemInfo_ptr_t, [size_t]),
                ("size",),
            ),
            NRT_MemInfo_new_varsize_dtor: (
                ir.FunctionType(MemInfo_ptr_t, [size_t, i8p]),
                ("size", "dtor"),
            ),
            NRT_MemInfo_varsize_free: (
                ir.FunctionType(void, [MemInfo_ptr_t, i8p]),
                ("mi", "ptr"),
            ),
            NRT_MemInfo_varsize_alloc: (
                ir.FunctionType(i8p, [MemInfo_ptr_t, size_t]),
                ("mi", "size"),
            ),
            NRT_MemInfo_varsize_realloc: (
                ir.FunctionType(i8p, [MemInfo_ptr_t, size_t]),
                ("mi", "size"),
            ),
            NRT_Reallocate: (ir.FunctionType(i8p, [i8p, size_t]), ("ptr", "size")),
            NRT_Reallocate_NoCopy: (
                ir.FunctionType(i8p, [i8p, size_t]),
                ("ptr", "size")),
            NRT_dealloc: (ir.FunctionType(void, [MemInfo_ptr_t]), ("mi",)),
            NRT_Free: (ir.FunctionType(void, [i8p]), ("ptr",)),
            NRT_MemInfo_destroy: (ir.FunctionType(void, [MemInfo_ptr_t]), ("mi",)),
            nrt_internal_custom_dtor: (
                ir.FunctionType(void, [i8p, size_t, i8p]),
                ("ptr", "size", "info"),
            ),
            nrt_allocate_meminfo_and_data: (
                ir.FunctionType(i8p, [size_t, i8pp, i8p]),
                ("size", "mi_out", "allocator"),
            ),
            allocate_varlen_buffer: (ir.FunctionType(i8p, [i64, i64]), ()),
            malloc: (ir.FunctionType(i8p, [size_t]), ("size",)),
            free: (ir.FunctionType(void, [i8p]), ("ptr",)),
            realloc: (ir.FunctionType(i8p, [i8p, size_t]), ()),
            # debug functions
            nrt_debug_incr: (ir.FunctionType(void, []), ()),
            nrt_debug_decr: (ir.FunctionType(void, []), ()),
            nrt_debug_spaces: (ir.FunctionType(void, []), ()),
        }

    def set_hashsecrets(self):
        for v in hashsecret.values():
            gv = ir.GlobalVariable(self.module, i64, v.symbol)
            gv.initializer = i64(v.value.value)

    def NRT_Debug(self, fmt, *args):
        # only debug if verbose is on
        if self.debug_nrt:
            builder = self.CURRENT_BUILDER
            msg = f"[NRT] {fmt}"
            if not msg.endswith("\n"):
                msg += "\n"

            gv = builder.load(self.module.get_global(nrt_global_var))
            cgutils.printf(builder, "%*d", gv, gv)
            cgutils.printf(builder, msg, *args)

    @contextmanager
    def nrt_debug_ctx(self):
        if self.debug_nrt:
            builder = self.CURRENT_BUILDER
            try:
                self.NRT_Debug(self.CURRENT_FUNCTION)
                builder.call(self.nrt_debug_incr, [])
                yield
            finally:
                builder.call(self.nrt_debug_decr, [])
        else:
            yield

    def define_nrt_debug_incr(self, builder, args):
        gv = self.module.get_global(nrt_global_var)
        builder.store(builder.add(builder.load(gv), i64(4)), gv)
        builder.ret_void()

    def define_nrt_debug_decr(self, builder, args):
        gv = self.module.get_global(nrt_global_var)
        builder.store(builder.sub(builder.load(gv), i64(4)), gv)
        builder.ret_void()

    def define_nrt_debug_spaces(self, builder, args):
        cgutils.printf(builder, "\t")
        builder.ret_void()

    def __getattribute__(self, __name):
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            if __name in self.function_types:
                return self._get_function(__name)
            raise

    def _get_function_builder(self, fn_name):
        fn = self._get_function(fn_name)
        block = fn.append_basic_block(name="entry")
        return ir.IRBuilder(block)

    def _get_from_meminfo(self, mi, kind):
        builder = self.CURRENT_BUILDER
        table = {
            "refct": REFCT_IDX,
            "dtor": DTOR_IDX,
            "dtor_info": DTOR_INFO_IDX,
            "data": DATA_IDX,
            "size": SIZE_IDX,
            "allocator": EXTERNAL_ALLOCATOR_IDX,
        }
        idx = table[kind]
        return builder.load(
            builder.gep(mi, [i32(0), i32(idx)], inbounds=True), name=kind
        )

    def _set_on_meminfo(self, mi, kind, val):
        builder = self.CURRENT_BUILDER
        table = {
            "refct": REFCT_IDX,
            "dtor": DTOR_IDX,
            "dtor_info": DTOR_INFO_IDX,
            "data": DATA_IDX,
            "size": SIZE_IDX,
            "allocator": EXTERNAL_ALLOCATOR_IDX,
        }
        idx = table[kind]
        return builder.store(val, builder.gep(mi, [i32(0), i32(idx)], inbounds=True))

    def _get_function(self, fn_name):
        fnty = self.function_types[fn_name][0]
        fn = cgutils.get_or_insert_function(self.module, fnty, fn_name)
        return fn

    def _get_function_arg_names(self, fn_name):
        return self.function_types[fn_name][1]

    def _declare_function(self, fn_name):
        fn = self._get_function(fn_name)
        argnames = self._get_function_arg_names(fn_name)
        if argnames is not None:
            assert isinstance(argnames, tuple), argnames
            assert len(argnames) == len(fn.args), (len(argnames), len(fn.args))
            for arg, name in zip(fn.args, argnames):
                arg.name = name
        if self.debug_nrt:
            fn.attributes.add("noinline")
        return fn

    def _ensure_impl_exists(self):
        # Ensure for each function in function_types, there is a corresponding
        # implementation
        skip_list = ("allocate_varlen_buffer", "realloc", "malloc", "free")
        for fn_name in self.function_types.keys():
            if fn_name not in skip_list:
                defn = f"define_{fn_name}"
                getattr(self, defn), defn

    def define(self):
        self._ensure_impl_exists()

        for name in dir(self):
            if not name.startswith("define_"):
                continue

            if name in _debug_functions and not self.debug_nrt:
                continue

            meth = getattr(self, name)
            fn_name = name[len("define_"):]
            self._declare_function(fn_name)
            builder = self._get_function_builder(fn_name)
            args = builder.function.args
            self.CURRENT_FUNCTION = fn_name
            self.CURRENT_BUILDER = builder
            meth(builder, args)

    def define_NRT_MemInfo_call_dtor(self, builder, args):
        with self.nrt_debug_ctx():
            [mi] = args
            self.NRT_Debug("NRT_MemInfo_call_dtor mi=%p\n", mi)

            dtor = self._get_from_meminfo(mi, "dtor")
            not_null = cgutils.is_not_null(builder, dtor)
            with cgutils.if_likely(builder, not_null):
                dtor_fn = builder.bitcast(dtor, self.NRT_DTOR_FUNCTION.as_pointer())
                data = self._get_from_meminfo(mi, "data")
                size = self._get_from_meminfo(mi, "size")
                dtor_info = self._get_from_meminfo(mi, "dtor_info")
                builder.call(dtor_fn, [data, size, dtor_info])

            builder.call(self.NRT_MemInfo_destroy, [mi])
        builder.ret_void()

    def define_NRT_incref(self, builder, args):
        # with self.nrt_debug_ctx():
        #     pass
        builder.ret_void()

    def define_NRT_decref(self, builder, args):
        # with self.nrt_debug_ctx():
        #     pass
        builder.ret_void()

    def define_NRT_MemInfo_alloc_safe_aligned(self, builder, args):
        # Just call the non-aligned version for now
        with self.nrt_debug_ctx():
            [size, align] = args
            ptr = builder.call(self.NRT_MemInfo_alloc_safe, [size])
            self.NRT_Debug("ptr=%p\n", ptr)
        builder.ret(ptr)

    def define_NRT_MemInfo_data_fast(self, builder, args):
        with self.nrt_debug_ctx():
            [ptr] = args
            self.NRT_Debug("ptr=%p\n", ptr)
            not_null = builder.icmp_signed("!=", ptr, NULL)
            with builder.if_else(not_null, likely=True) as (then, otherwise):
                with then:
                    bb_then = builder.basic_block
                    mi_ptr = builder.bitcast(
                        ptr, self.MemInfo_ptr_t, name="meminfo_ptr"
                    )
                    data = self._get_from_meminfo(mi_ptr, "data")
                    self.NRT_Debug("data: %p\n", data)
                with otherwise:
                    # there is a specific case where ptr is NULL and doesn't
                    # crash Numba
                    bb_else = builder.basic_block
                    data_null = NULL
            phi = builder.phi(i8p)
            phi.add_incoming(data, bb_then)
            phi.add_incoming(data_null, bb_else)
        builder.ret(phi)

    def define_NRT_Reallocate_NoCopy(self, builder, args):
        with self.nrt_debug_ctx():
            # Because the previously reallocated memory might be freed by
            # realloc, RBC cannot call it in this function. Therefore, instead
            # of reallocation, RBC allocates a new block of memory and returns
            # it to the caller. The caller is responsible for copying memory
            # from the old buffer into the most recent allocated one."
            [ptr, size] = args
            new_ptr = builder.call(self.NRT_Allocate_External, [size, NULL])
            self.NRT_Debug("ptr=%p size=%zu -> new_ptr=%p", ptr, size, new_ptr)
        builder.ret(new_ptr)

    def define_NRT_Reallocate(self, builder, args):
        with self.nrt_debug_ctx():
            # Because the previously reallocated memory might be freed by
            # realloc, RBC cannot call it in this function. Therefore, instead
            # of reallocation, RBC allocates a new block of memory and returns
            # it to the caller. The caller is responsible for copying memory
            # from the old buffer into the most recent allocated one."
            [ptr, size] = args
            new_ptr = builder.call(self.NRT_Allocate_External, [size, NULL])
            self.NRT_Debug("ptr=%p size=%zu -> new_ptr=%p", ptr, size, new_ptr)
            msg = ('The "NRT_Reallocate" function is a custom implementation '
                   'of the "realloc" function from the C standard library. '
                   'Unlike the original "realloc" function, this '
                   'implementation does not copy memory from the old buffer to '
                   'the new one. Instead, it simply allocates a new buffer '
                   'with the given size and returns a pointer to it. The old '
                   'buffer remains unchanged, and it is the responsibility of '
                   'the caller to copy data from the old buffer to the new one '
                   'if necessary.\n'
                   'See RBC issue #536')
            self.NRT_Debug(msg)
        builder.ret(new_ptr)

    def define_NRT_dealloc(self, builder, args):
        with self.nrt_debug_ctx():
            [mi] = args
            self.NRT_Debug("meminfo: %p\n", mi)
            ptr = builder.bitcast(mi, i8p)
            builder.call(self.NRT_Free, [ptr])
        builder.ret_void()

    def define_NRT_Free(self, builder, args):
        # this is a no-op function
        with self.nrt_debug_ctx():
            [ptr] = args
            self.NRT_Debug("# NRT_Free %p\n", ptr)
            # free is done automatically in HeavyDB
        builder.ret_void()

    def define_NRT_MemInfo_varsize_alloc(self, builder, args):
        with self.nrt_debug_ctx():
            [mi, size] = args

            data = builder.call(self.NRT_Allocate, [size])
            # TODO: check if data is NULL
            self._set_on_meminfo(mi, "data", data)
            self._set_on_meminfo(mi, "size", size)
            self.NRT_Debug("%p size=%zu -> data=%p\n", mi, size, data)
        builder.ret(data)

    def define_NRT_MemInfo_varsize_realloc(self, builder, args):
        with self.nrt_debug_ctx():
            [mi, size] = args
            data = self._get_from_meminfo(mi, "data")

            # Is this the only function that calls NRT_Reallocate?
            new_data = builder.call(self.NRT_Reallocate_NoCopy, [data, size],
                                    name="new_data")

            # RBC/NRT cannot use realloc as the old memory gets freed in the
            # process. Instead, memory is copied from the old buffer to the new
            # one here. The previous memory is freed upon UD[T]F return
            old_size = self._get_from_meminfo(mi, "size")
            cond = builder.icmp_signed("<", old_size, size)
            memcpy_size = builder.select(cond, old_size, size)

            self.NRT_Debug(
                "memcpy %zu bytes from data=%p --> new_data=%p",
                memcpy_size,
                data,
                new_data,
            )

            cgutils.raw_memcpy(
                builder, dst=new_data, src=data, count=memcpy_size, itemsize=i64(1)
            )

            self.NRT_Debug("new_data=%p - size=%zu\n", new_data, size)

            # TODO: check if new_data is NULL
            self._set_on_meminfo(mi, "data", new_data)
            self._set_on_meminfo(mi, "size", size)
        builder.ret(new_data)

    def define_NRT_MemInfo_varsize_free(self, builder, args):
        with self.nrt_debug_ctx():
            [mi, ptr] = args
            builder.call(self.NRT_Free, [ptr])

            mi_data = self._get_from_meminfo(mi, "data")
            eq = builder.icmp_signed("==", ptr, mi_data)
            with builder.if_then(eq):
                self._set_on_meminfo(mi, "data", NULL)
        builder.ret_void()

    def define_NRT_MemInfo_destroy(self, builder, args):
        with self.nrt_debug_ctx():
            [mi] = args
            self.NRT_Debug("mi=%p\n", mi)
            builder.call(self.NRT_dealloc, [mi])
        builder.ret_void()

    def define_nrt_internal_custom_dtor(self, builder, args):
        with self.nrt_debug_ctx():
            [ptr, size, info] = args

            dtor = builder.bitcast(info, self.NRT_DTOR_FUNCTION.as_pointer())
            self.NRT_Debug("ptr=%p, info=%p dtor=%p\n", ptr, info, dtor)

            not_null = cgutils.is_not_null(builder, dtor)
            with cgutils.if_likely(builder, not_null):
                info_ = NULL
                builder.call(dtor, [ptr, size, info_])
        builder.ret_void()

    def define_nrt_allocate_meminfo_and_data(self, builder, args):
        with self.nrt_debug_ctx():
            [size, mi_out, allocator] = args
            alloc_size = builder.add(size, self.sizeof_MemInfo_t)
            base = builder.call(self.NRT_Allocate_External, [alloc_size, allocator])
            builder.store(base, mi_out)
            out = builder.gep(base, [self.sizeof_MemInfo_t], inbounds=True)
            self.NRT_Debug("base=%p out=%p\n", base, out)
        builder.ret(out)

    def define_NRT_Allocate(self, builder, args):
        # allocator is always null as we don't use this argument in RBC/HeavyDB
        with self.nrt_debug_ctx():
            [size] = args
            allocator = NULL
            ret = builder.call(self.NRT_Allocate_External, [size, allocator])
        builder.ret(ret)

    def define_NRT_Allocate_External(self, builder, args):
        # allocator is always null as we don't use this argument in RBC/HeavyDB
        with self.nrt_debug_ctx():
            [size, allocator_] = args

            # ptr = builder.call(self.malloc, [size])
            # allocate_varlen_buffer calls malloc using the formula:
            #   malloc((elem_count + 1) * elem_size)
            elem_count = builder.sub(size, i64(1), name="element_count")
            elem_size = i64(1)
            ptr = builder.call(self.allocate_varlen_buffer, [elem_count, elem_size])
            self.NRT_Debug(
                "allocate_varlen_buffer=%p - element_count: %ld", ptr, elem_count
            )
            # self.NRT_Debug("malloc: bytes=%zu -> ptr=%p", size, ptr)
        builder.ret(ptr)

    def define_NRT_MemInfo_alloc_dtor(self, builder, args):
        with self.nrt_debug_ctx():
            mi = builder.alloca(self.MemInfo_ptr_t, name="mi")
            mi_cast = builder.bitcast(mi, i8pp, "mi_cast")
            [size, dtor] = args
            allocator = NULL
            data = builder.call(
                self.nrt_allocate_meminfo_and_data,
                [size, mi_cast, allocator],
                name="data",
            )

            self.NRT_Debug("data=%p size=%zu\n", data, size)
            # TODO: check if data is null
            dtor_function = builder.bitcast(
                self._get_function(nrt_internal_custom_dtor), i8p
            )
            dtor_info = builder.bitcast(dtor, i8p)
            allocator = NULL
            builder.call(
                self.NRT_MemInfo_init,
                [builder.load(mi), data, size, dtor_function, dtor_info, allocator],
            )
        builder.ret(builder.load(mi))

    def define_NRT_MemInfo_alloc(self, builder, args):
        with self.nrt_debug_ctx():
            mi = builder.alloca(self.MemInfo_ptr_t, name="mi")
            mi_cast = builder.bitcast(mi, i8pp, name="mi_cast")
            [size] = args
            allocator = NULL
            data = builder.call(
                self.nrt_allocate_meminfo_and_data,
                [size, mi_cast, allocator],
                name="data",
            )
            # TODO: check if data is null

            self.NRT_Debug("%p\n", data)
            dtor = NULL
            dtor_info = NULL
            allocator = NULL
            builder.call(
                self.NRT_MemInfo_init,
                [builder.load(mi), data, size, dtor, dtor_info, allocator],
            )
        builder.ret(builder.load(mi))

    def define_NRT_MemInfo_new_varsize(self, builder, args):
        with self.nrt_debug_ctx():
            [size] = args
            self.NRT_Debug("size=%zu\n", size)

            # TODO: check if data is null
            data = builder.call(self.NRT_Allocate, [size])
            mi = builder.call(self.NRT_MemInfo_new, [data, size, NULL, NULL], name="mi")
            self.NRT_Debug("size=%zu -> meminfo=%p, data=%p\n", size, mi, data)
        builder.ret(mi)

    def define_NRT_MemInfo_new(self, builder, args):
        with self.nrt_debug_ctx():
            [data, size, dtor, dtor_info] = args
            ptr = builder.call(self.NRT_Allocate, [self.sizeof_MemInfo_t])
            mi = builder.bitcast(ptr, self.MemInfo_ptr_t, name="mi")
            self.NRT_Debug("ptr=%p\n", ptr)

            is_not_null = cgutils.is_not_null(builder, mi)
            with cgutils.if_likely(builder, is_not_null):
                self.NRT_Debug(
                    "mi=%p data=%p size=%zu dtor=%p dtor_info=%p\n",
                    mi,
                    data,
                    size,
                    dtor,
                    dtor_info,
                )
                builder.call(
                    self.NRT_MemInfo_init, [mi, data, size, dtor, dtor_info, NULL]
                )
        builder.ret(mi)

    def define_NRT_MemInfo_new_varsize_dtor(self, builder, args):
        with self.nrt_debug_ctx():
            [size, dtor] = args
            self.NRT_Debug("size: %d\n", size)

            mi = builder.call(self.NRT_MemInfo_new_varsize, [size])
            self.NRT_Debug("mi=%p\n", mi)

            is_not_null = cgutils.is_not_null(builder, mi)
            with cgutils.if_likely(builder, is_not_null):
                self._set_on_meminfo(mi, "dtor", dtor)
        builder.ret(mi)

    def define_NRT_MemInfo_alloc_safe(self, builder, args):
        with self.nrt_debug_ctx():
            [size] = args
            self.NRT_Debug("NRT_MemInfo_alloc_safe -> NRT_MemInfo_alloc_dtor_safe\n")
            NRT_dtor_function = NULL
            ret = builder.call(
                self.NRT_MemInfo_alloc_dtor_safe, [size, NRT_dtor_function]
            )
        builder.ret(ret)

    def define_NRT_MemInfo_alloc_dtor_safe(self, builder, args):
        with self.nrt_debug_ctx():
            # wire implementation to unsafe version
            [size, dtor] = args
            data = builder.call(self.NRT_MemInfo_alloc_dtor, [size, dtor], name="data")
            self.NRT_Debug("data=%p size=%zu\n", data, size)
        builder.ret(data)

    def define_NRT_MemInfo_init(self, builder, args):
        with self.nrt_debug_ctx():
            [mi, data, size, dtor, dtor_info, external_allocator] = args
            self._set_on_meminfo(mi, "refct", self.size_t(1))
            self._set_on_meminfo(mi, "dtor", dtor)
            self._set_on_meminfo(mi, "dtor_info", dtor_info)
            self._set_on_meminfo(mi, "data", data)
            self._set_on_meminfo(mi, "size", size)
            self._set_on_meminfo(mi, "allocator", external_allocator)
            self.NRT_Debug("mi=%p data=%p size=%zu\n", mi, data, size)
        builder.ret_void()


def create_nrt_functions(target_context, debug):
    nrt = RBC_NRT(target_context, verbose=debug)
    return nrt.module




'''RBC Point2D type.
'''

__all__ = ['Point2D']

from llvmlite import ir
from numba.core import cgutils, extending
from numba.core import types as nb_types

from rbc import typesystem

from .metatype import HeavyDBMetaType

i32 = ir.IntType(32)
i64 = ir.IntType(64)


class Point2D(metaclass=HeavyDBMetaType):
    """

    A Point2D struct is just a wrapper for a pair of doubles

    ... code-block:: c

        {
            double x;
            double y;
        }
    """

    def __init__(self, x: float, y: float) -> None:
        pass

    @property
    def x(self) -> float:
        pass

    @property
    def y(self) -> float:
        pass


class Point2DNumbaType(nb_types.Type):
    pass


class HeavyDBPoint2D(typesystem.Type):

    @property
    def custom_params(self):
        return {
            'name': 'Point2D',
            'NumbaType': Point2DNumbaType,
        }

    def tonumba(self, bool_is_int8=None):
        x = typesystem.Type.fromstring('double x')
        y = typesystem.Type.fromstring('double y')
        point2d = typesystem.Type(x, y).params(**self.custom_params)
        return point2d.tonumba(bool_is_int8=True)


@extending.type_callable(Point2D)
def type_heavydb_point2d(context):
    def typer(x, y):
        if isinstance(x, nb_types.Float) and isinstance(y, nb_types.Float):
            return typesystem.Type.fromobject('Point2D').tonumba()
    return typer


@extending.lower_builtin(Point2D, nb_types.Float, nb_types.Float)
def heavydb_point2d_ctor(context, builder, sig, args):
    [x, y] = args
    typ = sig.return_type
    point = cgutils.create_struct_proxy(typ)(context, builder)
    point.x = x
    point.y = y
    return point._getvalue()


@extending.intrinsic
def intrinsic_point2d_get(typingctx, point2d, pos):
    sig = nb_types.double(point2d, pos)

    def codegen(context, builder, sig, args):
        [arg, pos] = args
        assert isinstance(pos, ir.Constant)
        return builder.extract_value(arg, [pos.constant])

    return sig, codegen


@extending.overload_attribute(Point2DNumbaType, 'x')
def impl_poind2d_attr_x(point2d):
    def impl(point2d):
        return intrinsic_point2d_get(point2d, 0)
    return impl


@extending.overload_attribute(Point2DNumbaType, 'y')
def impl_poind2d_attr_y(point2d):
    def impl(point2d):
        return intrinsic_point2d_get(point2d, 1)
    return impl



from rbc import heavydb


def test_column_geo_rewire():
    # ensure column rewiring works
    column_geo = set([t for t in dir(heavydb) if t.startswith('Geo')])
    expected_set = set(['GeoPoint', 'GeoMultiPoint',
                        'GeoLineString', 'GeoMultiLineString',
                        'GeoPolygon', 'GeoMultiPolygon'])
    assert len(column_geo) == len(expected_set)
    assert column_geo == expected_set



import pytest
import numpy as np
from rbc.heavydb import Point2D
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(6, 4),
                             suffices=['point']):
        define(o)
        yield o


def define(heavydb):
    @heavydb("int32(TableFunctionManager, Column<GeoPoint>, OutputColumn<Z>, OutputColumn<Z>)",
             Z=['double'], devices=['cpu'])
    def rbc_ct_coords(mgr, points, xcoords, ycoords):
        size = len(points)
        mgr.set_output_row_size(size)
        for i in range(size):
            if points.is_null(i):
                xcoords.set_null(i)
                ycoords.set_null(i)
            else:
                point = points[i]
                xcoords[i] = point.x
                ycoords[i] = point.y
        return size

    @heavydb("int32(TableFunctionManager, Column<T>, double, double, OutputColumn<T>)",
             T=['GeoPoint'], devices=['cpu'])
    def rbc_ct_shift(mgr, points, x, y, shifted_points):
        size = len(points)
        mgr.set_output_row_size(size)
        for i in range(size):
            if points.is_null(i):
                shifted_points.set_null(i)
            else:
                point = points[i]
                shifted_points.set_item(i, Point2D(point.x + x, point.y + y))
        return size

    @heavydb("int32(TableFunctionManager, Cursor<Column<T>, Column<T>>, T, T, OutputColumn<Z>)",
             T=['double'], Z=['GeoLineString'], devices=['cpu'])
    def rbc_ct_make_linestring2(mgr, x, y, dx, dy, linestrings):
        size = len(x)
        mgr.set_output_item_values_total_number(0, size * 4)
        mgr.set_output_row_size(size)
        for i in range(size):
            if x.is_null(i) or y.is_null(i):
                linestrings.set_null(i)
            else:
                line = [x[i], y[i], x[i] + dx, y[i] + dy]
                linestrings[i].from_coords(line)
        return size


@pytest.mark.parametrize('suffix', ('point',))
@pytest.mark.parametrize('col', ('p1', 'p2', 'p3', 'p4'))
def test_ct_coords(heavydb, suffix, col):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    query = (f'select * from table(rbc_ct_coords(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_coords(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('suffix', ('point',))
@pytest.mark.parametrize('col', ('p1', 'p2', 'p3', 'p4'))
def test_ct_shift(heavydb, suffix, col):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    query = (f'select * from table(rbc_ct_shift(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), 1.5, -2.5));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_shift(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), 1.5, -2.5));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('col', ('p1', 'p2', 'p3', 'p4'))
@pytest.mark.parametrize('dx,dy', [('2', '3')])
def test_ct_make_linestring2(heavydb, col, dx, dy):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    suffix = 'point'
    table = heavydb.table_name + suffix

    query = ("SELECT * FROM TABLE(rbc_ct_make_linestring2(CURSOR(SELECT "
             f"ST_X({col}), ST_Y({col}) FROM {table}), {dx}, {dy}));")
    _, result = heavydb.sql_execute(query)

    query = (f"SELECT * FROM TABLE(ct_make_linestring2(CURSOR(SELECT "
             f"ST_X({col}), ST_Y({col}) FROM {table}), {dx}, {dy}));")
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)



import inspect
from rbc.targetinfo import TargetInfo
from numba.core import funcdesc


def gen_codegen(fn_name):
    def codegen(context, builder, sig, args):
        # Need to retrieve the function name again
        fndesc = funcdesc.ExternalFunctionDescriptor(fn_name, sig.return_type, sig.args)
        func = context.declare_external_function(builder.module, fndesc)
        return builder.call(func, args)

    return codegen


def dispatch_codegen(cpu, gpu):
    def inner(context, builder, sig, args):
        impl = cpu if TargetInfo().is_cpu else gpu
        return impl(context, builder, sig, args)

    return inner


def sanitize(name):
    forbidden_names = ('in')
    if name in forbidden_names:
        return f"{name}_"
    return name


def make_intrinsic(fname, retty, argnames, module_name, module_globals, doc):
    argnames = tuple(map(lambda x: sanitize(x), argnames))
    fn_str = f'''
from numba.core.extending import intrinsic
@intrinsic
def {fname}(typingctx, {", ".join(argnames)}):
    from rbc.typesystem import Type
    retty_ = Type.fromobject("{retty}").tonumba()
    argnames_ = tuple(map(lambda x: Type.fromobject(x).tonumba(), [{", ".join(argnames)}]))  # noqa: E501
    signature = retty_(*argnames_)
    def codegen(context, builder, sig, args):
        from numba.core import funcdesc
        fndesc = funcdesc.ExternalFunctionDescriptor("{fname}", sig.return_type, sig.args)  # noqa: E501
        func = context.declare_external_function(builder.module, fndesc)
        return builder.call(func, args)
    return signature, codegen
'''

    exec(fn_str, module_globals)
    fn = module_globals[fname]
    fn.__doc__ = doc

    # https://chriswarrick.com/blog/2018/09/20/python-hackery-merging-signatures-of-two-python-functions/
    # this is to remove `typingctx` from showing up in the docs
    sig = inspect.signature(fn)
    assert list(sig.parameters.keys())[0] == 'typingctx'
    fn.__signature__ = sig.replace(parameters=tuple(sig.parameters.values())[1:])
    return fn



import pytest
import numpy as np
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(6, 4),
                             suffices=['multilinestring']):
        define(o)
        yield o


def define(heavydb):
    @heavydb("int32(TableFunctionManager, Column<Z>, OutputColumn<K>)",
             Z=['GeoMultiLineString'], K=['GeoPolygon'], devices=['cpu'])
    def rbc_ct_to_polygon(mgr, mlinestrings, polygons):
        size = len(mlinestrings)
        mgr.set_output_item_values_total_number(0, mlinestrings.get_n_of_values())
        mgr.set_output_row_size(size)
        # Initialize polygons
        for i in range(size):
            if mlinestrings.is_null(i):
                polygons.set_null(i)
            else:
                coords = mlinestrings[i].to_coords()
                polygons[i].from_coords(coords)
        return size


@pytest.mark.parametrize('col', ('ml1', 'ml2', 'ml3', 'ml4'))
def test_ct_to_polygon(heavydb, col):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    suffix = 'multilinestring'
    func = 'ct_to_polygon'

    query = (f'select * from table(rbc_{func}(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table({func}(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)



import pytest
import numpy as np
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(6, 4),
                             suffices=['polygon']):
        define(o)
        yield o


def define(heavydb):
    @heavydb("int32(TableFunctionManager, Column<Z>, int64_t, OutputColumn<K>)",
             Z=['GeoPolygon'], K=['GeoLineString'], devices=['cpu'])
    def rbc_ct_linestringn(mgr, polygons, n, linestrings):
        size = len(polygons)
        mgr.set_output_item_values_total_number(0, polygons.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if polygons.is_null(i):
                linestrings.set_null(i)
            else:
                sz = len(polygons[i])
                if n < 1 or n > sz:
                    linestrings.set_null(i)
                else:
                    poly = polygons[i]
                    ring = poly[n - 1]
                    linestrings[i] = ring
        return size

    @heavydb("int32(TableFunctionManager, Column<Z>, OutputColumn<K>)",
             Z=['GeoPolygon'], K=['GeoMultiLineString'], devices=['cpu'])
    def rbc_ct_to_multilinestring(mgr, polygons, mlinestrings):
        size = len(polygons)
        mgr.set_output_item_values_total_number(0, polygons.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if polygons.is_null(i):
                mlinestrings.set_null(i)
            else:
                polygon_coords = polygons[i].to_coords()
                mlinestrings[i].from_coords(polygon_coords)
        return size

    @heavydb("int32(TableFunctionManager, Column<Z>, OutputColumn<K>)",
             Z=['GeoPolygon'], K=['GeoMultiPolygon'], devices=['cpu'])
    def rbc_ct_make_multipolygon(mgr, polygons, mpolygons):
        size = len(polygons)
        mgr.set_output_item_values_total_number(0, polygons.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if polygons.is_null(i):
                mpolygons.set_null(i)
            else:
                polygon_coords = polygons[i].to_coords()
                mpolygons[i].from_coords([polygon_coords])
        return size


@pytest.mark.parametrize('n', ('1', '2', '3'))
@pytest.mark.parametrize('col', ('p1', 'p2', 'p3', 'p4'))
def test_ct_coords(heavydb, n, col):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    suffix = 'polygon'

    query = (f'select * from table(rbc_ct_linestringn(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), {n}));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_linestringn(cursor(select {col} '
             f'from {heavydb.table_name}{suffix}), {n}));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('col', ('p1', 'p2', 'p3', 'p4'))
def test_ct_to_multilinestring(heavydb, col):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    suffix = 'polygon'

    query = (f'select * from table(rbc_ct_to_multilinestring(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_to_multilinestring(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('col', ('p1', 'p2', 'p3', 'p4'))
def test_ct_make_multipolygon(heavydb, col):
    if heavydb.version[:2] < (6, 4):
        pytest.skip('Requires HeavyDB 6.4 or newer')

    suffix = 'polygon'

    query = (f'select * from table(rbc_ct_make_multipolygon(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, result = heavydb.sql_execute(query)

    query = (f'select * from table(ct_make_multipolygon(cursor(select {col} '
             f'from {heavydb.table_name}{suffix})));')
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)




from .server import Server                # noqa: F401
from .client import Client                # noqa: F401
from .dispatcher import Dispatcher        # noqa: F401
from .types import Buffer, NDArray, Data  # noqa: F401
from .utils import dispatchermethod       # noqa: F401



from rbc.targetinfo import TargetInfo


def test_basic():
    host = TargetInfo.host()
    assert host.name == 'host_cpu'
    with host:
        ti = TargetInfo()
        assert ti is host


def test_to_from_dict():
    host = TargetInfo.host()
    d = host.todict()
    host2 = TargetInfo.fromdict(d)
    assert host.__dict__ == host2.__dict__


def test_host_cache():
    host = TargetInfo.host()
    host2 = TargetInfo.host()
    assert host is host2



"""
Array API specification for data type functions.

https://data-apis.org/array-api/latest/API_specification/data_type_functions.html
"""
from rbc.stdlib import Expose

__all__ = [
    "astype",
    "broadcast_arrays",
    "broadcast_to",
    "can_cast",
    "finfo",
    "iinfo",
    "result_type",
]

expose = Expose(globals(), "data_type_functions")


@expose.not_implemented("astype")
def _array_api_astype(x, dtype, *, copy=True):
    """
    Copies an array to a specified data type irrespective of
    `Type Promotion Rules <https://data-apis.org/array-api/latest/API_specification/type_promotion.html#type-promotion>`_ rules.
    """
    pass


@expose.not_implemented("broadcast_arrays")
def _array_api_broadcast_arrays(*arrays):
    """
    Broadcasts one or more arrays against one another.
    """
    pass


@expose.not_implemented("broadcast_to")
def _array_api_broadcast_to(x, shape):
    """
    Broadcasts an array to a specified shape.
    """
    pass


@expose.not_implemented("can_cast")
def _array_api_can_cast(from_, to):
    """
    Determines if one data type can be cast to another data type according
    `Type Promotion Rules <https://data-apis.org/array-api/latest/API_specification/type_promotion.html#type-promotion>`_ rules.
    """
    pass


@expose.not_implemented("finfo")
def _array_api_finfo(type):
    """
    Machine limits for floating-point data types.
    """
    pass


@expose.not_implemented("iinfo")
def _array_api_iinfo(type):
    """
    Machine limits for integer data types.
    """
    pass


@expose.not_implemented("result_type")
def _array_api_result_type(*arrays_and_dtypes):
    """
    Returns the dtype that results from applying the type promotion
    rules (see `Type Promotion Rules <https://data-apis.org/array-api/latest/API_specification/type_promotion.html#type-promotion>`_)
    to the arguments.
    """
    pass



import pytest
from rbc.stdlib import array_api
from rbc.tests import heavydb_fixture
from numba import TypingError


@pytest.fixture(scope="module")
def heavydb():

    for o in heavydb_fixture(globals(), load_test_data=False):
        yield o


unsupported_functions = [
    # statistical_function
    'std', 'var',
    # manipulation_functions
    "concat", "expand_dims", "flip", "permute_dims", "reshape", "roll", "squeeze", "stack",
    # utility_functions
    "all", "any",
    # sorting_functions
    "argsort", "sort",
    # searching_functions
    "argmax", "argmin", "nonzero", "where",
    # elementwise_functions
    'float_power', 'divmod', 'cbrt', 'isnat',
    # set_functions
    "unique_all", "unique_counts", "unique_inverse", "unique_values",
    # linear_algebra_functions
    "matmul", "matrix_transpose", "tensordot", "vecdot",
    # data_type_functions
    "astype", "broadcast_arrays", "broadcast_to", "can_cast", "finfo", "iinfo", "result_type",
    # creation_functions
    'arange', 'asarray', 'eye', 'from_dlpack', 'linspace', 'meshgrid', 'tril', 'triu',
]


# ensure unimplemented functions raise a meaninful exception
@pytest.mark.parametrize('func_name', unsupported_functions)
def test_unimplemented(heavydb, func_name):

    func = getattr(array_api, func_name)

    @heavydb('int64(int64)')
    def test_exception_raised(x):
        return func(x)

    # NumbaNotSupportedError is captured and a TypingError is returned instead
    with pytest.raises(TypingError, match=f'Function "{func_name}" is not supported.'):
        heavydb.register()
    heavydb.unregister()



import sys
import atexit
import pytest
from rbc.tests import heavydb_fixture
from rbc.remotejit import RemoteJIT
from rbc.externals import cmath
from rbc.typesystem import Type
import numpy as np
import math


@pytest.fixture(scope="module")
def rjit(request):
    local = False
    rjit = RemoteJIT(debug=not True, local=local, port=11542)
    if not local:
        rjit.start_server(background=True)
        request.addfinalizer(rjit.stop_server)
        atexit.register(rjit.stop_server)
    define(rjit)
    return rjit


@pytest.fixture(scope="module")
def ljit(request):
    ljit = RemoteJIT(debug=not True, local=True, port=11543)
    define(ljit)
    return ljit


@pytest.fixture(scope="module")
def heavydb():

    for o in heavydb_fixture(globals()):
        define(o)
        yield o


cmath_funcs = (
    # Trigonometric
    ("cos", "float64 cos(float64)"),
    ("sin", "float64 sin(float64)"),
    ("tan", "float64 tan(float64)"),
    ("acos", "float64 acos(float64)"),
    ("asin", "float64 asin(float64)"),
    ("atan", "float64 atan(float64)"),
    ("atan2", "float64 atan2(float64, float64)"),
    # Hyperbolic
    ("cosh", "float64 cosh(float64)"),
    ("sinh", "float64 sinh(float64)"),
    ("tanh", "float64 tanh(float64)"),
    ("acosh", "float64 acosh(float64)"),
    ("asinh", "float64 asinh(float64)"),
    ("atanh", "float64 atanh(float64)"),
    # Exponential and logarithmic functions
    ("exp", "float64 exp(float64)"),
    ("frexp", "float64 frexp(float64)"),
    ("ldexp", "float64 ldexp(float64, int64)"),
    ("log", "float64 log(float64)"),
    ("log10", "float64 log10(float64)"),
    ("modf", "float64 modf(float64, float64*)"),
    ("exp2", "float64 exp2(float64)"),
    ("expm1", "float64 expm1(float64)"),
    ("ilogb", "float64 ilogb(float64)"),
    ("log1p", "float64 log1p(float64)"),
    ("log2", "float64 log2(float64)"),
    ("logb", "float64 logb(float64)"),
    # power functions
    ("pow", "float64 pow(float64, float64)"),
    ("sqrt", "float64 sqrt(float64)"),
    ("cbrt", "float64 cbrt(float64)"),
    ("hypot", "float64 hypot(float64, float64)"),
    # error and gamma functions
    ("erf", "float64 erf(float64)"),
    ("erfc", "float64 erfc(float64)"),
    ("tgamma", "float64 tgamma(float64)"),
    ("lgamma", "float64 lgamma(float64)"),
    # Rounding
    ("ceil", "float64 ceil(float64)"),
    ("floor", "float64 floor(float64)"),
    ("fmod", "float64 fmod(float64, float64)"),
    ("trunc", "float64 trunc(float64)"),
    ("round", "float64 round(float64)"),
    ("lround", "int64 lround(float64)"),
    ("llround", "int64 llround(float64)"),
    ("rint", "float64 rint(float64)"),
    ("lrint", "int64 lrint(float64)"),
    ("llrint", "int64 llrint(float64)"),
    ("nearbyint", "float64 nearbyint(float64)"),
    ("remainder", "float64 remainder(float64, float64)"),
    # Floating-point manipulation
    ("copysign", "float64 copysign(float64, float64)"),
    ("nan", "float64 nan(const char*)"),
    ("nextafter", "float64 nextafter(float64, float64)"),
    ("nexttoward", "float64 nexttoward(float64, float64)"),
    # Minimum, maximum, difference functions
    ("fdim", "float64 fdim(float64, float64)"),
    ("fmax", "float64 fmax(float64, float64)"),
    ("fmin", "float64 fmin(float64, float64)"),
    # Other functions
    ("fabs", "float64 fabs(float64)"),
    ("abs", "int64 abs(int64)"),
    ("fma", "float64 fma(float64, float64, float64)"),
)


funcs = {}


def _get_query_func(fname):
    PREFIX = "test_cmath"
    return f"{PREFIX}_{fname}"


def define(jit):
    global funcs

    def inner(cmath_func, query_func, signature):
        cmath_fn = getattr(cmath, cmath_func)
        t = Type.fromstring(signature)
        retty = str(t[0])
        argtypes = tuple(map(str, t[1]))
        arity = len(argtypes)

        # define callable
        if arity == 1:
            def fn(a):
                return cmath_fn(a)

        elif arity == 2:
            def fn(a, b):
                return cmath_fn(a, b)

        else:
            def fn(a, b, c):
                return cmath_fn(a, b, c)

        fn.__name__ = query_func
        fn = jit(f"{retty}({', '.join(argtypes)})", devices=["cpu"])(fn)
        return fn

    blocklist = ("frexp", "nan")

    for fname, signature in cmath_funcs:
        if fname not in blocklist:
            query_func = _get_query_func(fname)
            fn = inner(fname, query_func, signature)
            funcs[(jit, query_func)] = fn


def _get_pyfunc(fname):
    remap = {
        "acos": "arccos",
        "asin": "arcsin",
        "atan": "arctan",
        "atan2": "arctan2",
        "acosh": "arccosh",
        "asinh": "arcsinh",
        "atanh": "arctanh",
        "pow": "power",
        "tgamma": "gamma",
        "lround": "round",
        "llround": "round",
        "lrint": "round",
        "llrint": "round",
        "nearbyint": "round",
        "nexttoward": "nextafter",
        "fdim": "subtract",
    }

    if fname == "fma":
        return lambda a, b, c: a*b + c
    if fname == "ilogb":
        return lambda a: np.trunc(np.log2(a))
    if fname == "logb":
        return np.log2

    mod = math if fname in ["erf", "erfc", "tgamma", "lgamma"] else np
    fn = getattr(mod, remap.get(fname, fname))
    return fn


@pytest.mark.parametrize("fname,sig", cmath_funcs, ids=[item[0] for item in cmath_funcs])
def test_external_cmath_heavydb(heavydb, fname, sig):

    if fname in ["logb", "ilogb", "modf"]:
        pytest.skip(f"cmath function {fname} not supported")

    if fname in ["frexp", "nan"]:
        pytest.xfail(f"cmath function {fname} crashes heavydb server")

    if fname in ["remainder"]:
        pytest.xfail(f"cmath.{fname} wrong output!")

    table = heavydb.table_name
    query_func = _get_query_func(fname)
    pyfunc = _get_pyfunc(fname)

    if fname in ["acos", "asin", "atan"]:
        query = f"SELECT f8/10.0, {query_func}(f8/10.0) from {table}"
    elif fname in ["atan2"]:
        query = f"SELECT f8/10.0, f8/8.0, {query_func}(f8/10.0, f8/8.0) FROM {table}"
    elif fname in [
        "pow",
        "hypot",
        "fmod",
        "remainder",
        "nextafter",
        "nexttoward",
        "fdim",
        "fmax",
        "fmin",
    ]:
        query = f"SELECT f8+10.0, f8+1.0, {query_func}(f8+10.0, f8+1.0) FROM {table}"
    elif fname == "copysign":
        query = f"SELECT f8, -1*f8, {query_func}(f8, -1*f8) FROM {table}"
    elif fname == "fma":
        query = f"SELECT f8, f8, f8, {query_func}(f8, f8, f8) FROM {table}"
    elif fname == "ldexp":
        query = f"SELECT f8+1.0, 2, {query_func}(f8+1.0, 2) FROM {table}"
    elif fname == "atanh":
        query = f"SELECT f8/8.0, {query_func}(f8/8.0) from {table}"
    elif fname == "abs":
        query = f"SELECT -1*i8, {query_func}(-1*i8) from {table}"
    else:
        query = f"SELECT f8+10.0, {query_func}(f8+10.0) from {table}"

    _, result = heavydb.sql_execute(query)

    for values in result:
        if len(values) == 3:
            a, b, r = values
            assert np.isclose(r, pyfunc(a, b)), fname
        elif len(values) == 4:
            a, b, c, r = values
            assert np.isclose(r, pyfunc(a, b, c)), fname
        else:
            a, r = values
            assert np.isclose(r, pyfunc(a)), fname


@pytest.fixture(scope="module")
def input_data():
    return {
        'f8': np.arange(5, dtype=np.float64),
        'i8': np.arange(5, dtype=np.int64)
    }


@pytest.mark.parametrize("location", ['local', 'remote'])
@pytest.mark.parametrize("fname,sig", cmath_funcs, ids=[item[0] for item in cmath_funcs])
def test_external_cmath_remotejit(input_data, location, ljit, rjit, fname, sig):

    if fname in ["modf", "frexp", "nan"]:
        pytest.skip(f"cmath function {fname} requires a pointer argument")

    if fname in ["remainder", "ilogb", "logb"]:
        pytest.xfail(f"cmath function {fname} returns the wrong value")

    if fname in ["nexttoward"] and sys.platform == "darwin":
        pytest.xfail(f"{fname} fails on {sys.platform}")

    jit = rjit if location == 'remote' else ljit
    query_func = _get_query_func(fname)
    fn = funcs.get((jit, query_func))

    i8 = input_data['i8']
    f8 = input_data['f8']

    pyfunc = _get_pyfunc(fname)

    if fname in ["acos", "asin", "atan"]:
        args = (f8/10.0,)
    elif fname in ["logb"]:
        args = ((f8+1)*10.0,)
    elif fname in ["atan2"]:
        args = (f8/10.0, f8/8.0)
    elif fname in ["pow", "hypot", "fmod", "remainder", "nextafter",
                   "nexttoward", "fdim", "fmax", "fmin"]:
        args = (f8+10.0, f8+1.0)
    elif fname == "copysign":
        args = (f8, -1*f8)
    elif fname == "fma":
        args = (f8, f8, f8)
    elif fname == "ldexp":
        args = f8+1.0, np.full_like(f8, 2, dtype=np.int32)
    elif fname == "atanh":
        args = (f8/8.0,)
    elif fname == "abs":
        args = (-1*i8,)
    else:
        args = (f8 + 10.0,)

    result = list(map(lambda inputs: fn(*inputs), zip(*args)))

    for values in zip(*args, result):
        if len(values) == 3:
            a, b, r = values
            assert np.isclose(r, pyfunc(a, b)), fname
        elif len(values) == 4:
            a, b, c, r = values
            assert np.isclose(r, a * b + c), fname
        else:
            a, r = values
            assert np.isclose(r, pyfunc(a)), fname



__all__ = ['HeavyDBRowFunctionManagerType', 'RowFunctionManager']


from numba.core import extending, typing
from numba.core import types as nb_types

from rbc.errors import UnsupportedError
from rbc.external import external
from rbc.targetinfo import TargetInfo

from . import string_dict_proxy, text_encoding_none
from .metatype import HeavyDBMetaType
from .opaque_pointer import HeavyDBOpaquePtr, OpaquePtrNumbaType
from .utils import as_voidptr, global_str_constant


class HeavyDBRowFunctionManagerType(HeavyDBOpaquePtr):
    """RowFunctionManager<> is a typesystem custom type that
    represents a class type with the following public interface:

      struct RowFunctionManager { }
    """

    @property
    def numba_type(self):
        return RowFunctionManagerNumbaType

    @property
    def type_name(self):
        return "RowFunctionManager"


class RowFunctionManagerNumbaType(OpaquePtrNumbaType):
    pass


class RowFunctionManager(metaclass=HeavyDBMetaType):
    """
    TableFunctionManager is available in HeavyDB 6.2 or newer
    """

    TRANSIENT_DICT_DB_ID = 0
    TRANSIENT_DICT_ID = 0

    def getString(self, db_id: int, dict_id: int, str_arg: int) -> 'text_encoding_none.TextEncodingNone':  # noqa: E501
        """
        """

    def getStringDictionaryProxy(self, db_id: int, dict_id: int) -> 'string_dict_proxy.StringDictionaryProxy':  # noqa: E501
        """
        """

    def getDictDbId(self, func_name: str, arg_idx: int) -> int:
        """
        """

    def getDictId(self, func_name: str, arg_idx: int) -> int:
        """
        """


error_msg = 'RowFunctionManager is only available in HeavyDB 6.2 or newer (got %s)'


@extending.overload_method(RowFunctionManagerNumbaType, 'getDictId')
def heavydb_udf_manager_get_dict_id(mgr, func_name, arg_idx):
    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 2, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    defn = 'int32 RowFunctionManager_getDictId(int8*, int8*, int32)'
    get_dict_id_ = external(defn)

    def impl(mgr, func_name, arg_idx):
        func_name_ = global_str_constant("row_mgr_func_name", func_name)
        return get_dict_id_(as_voidptr(mgr), func_name_, arg_idx)
    return impl


@extending.overload_method(RowFunctionManagerNumbaType, 'getDictDbId')
def heavydb_udf_manager_get_db_id(mgr, func_name, arg_idx):
    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 2, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    defn = 'int32 RowFunctionManager_getDictDbId(int8*, int8*, int32)'
    get_dict_db_id_ = external(defn)

    def impl(mgr, func_name, arg_idx):
        func_name_ = global_str_constant("row_mgr_func_name", func_name)
        return get_dict_db_id_(as_voidptr(mgr), func_name_, arg_idx)
    return impl


@extending.overload_method(RowFunctionManagerNumbaType, 'getStringDictionaryProxy')
def heavydb_udf_manager_get_string_dict_proxy(mgr, db_id, dict_id):
    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 2, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    i8p = nb_types.CPointer(nb_types.int8)
    i32 = nb_types.int32
    proxy = string_dict_proxy.StringDictionaryProxyNumbaType()
    sig = typing.signature(proxy, i8p, i32, i32)

    symbol = 'RowFunctionManager_getStringDictionaryProxy'
    # TODO: define get_string_dict_proxy_ via external signature
    get_string_dict_proxy_ = nb_types.ExternalFunction(symbol, sig)

    def impl(mgr, db_id, dict_id):
        return get_string_dict_proxy_(as_voidptr(mgr), db_id, dict_id)
    return impl


@extending.overload_method(RowFunctionManagerNumbaType, 'getOrAddTransient')
def heavydb_udf_manager_get_or_add_transient(mgr, db_id, dict_id, str_arg):
    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 2, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    def impl(mgr, db_id, dict_id, str_arg):
        proxy = mgr.getStringDictionaryProxy(db_id, dict_id)
        return proxy.getOrAddTransient(str_arg)
    return impl


@extending.overload_method(RowFunctionManagerNumbaType, 'getString')
def heavydb_udf_manager_get_string(mgr, db_id, dict_id, string_id):
    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 2, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    def impl(mgr, db_id, dict_id, string_id):
        proxy = mgr.getStringDictionaryProxy(db_id, dict_id)
        return proxy.getString(string_id)
    return impl


@extending.overload_attribute(RowFunctionManagerNumbaType, 'TRANSIENT_DICT_DB_ID')
@extending.overload_attribute(RowFunctionManagerNumbaType, 'TRANSIENT_DICT_ID')
def heavydb_udf_manager_transient_dict_id(mgr):
    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 2, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    def impl(mgr):
        return 0
    return impl



from rbc.tests import heavydb_fixture
from rbc.heavydb import Array
import pytest

rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(),
                             suffices=['array', 'arraynull']):
        define(o)
        yield o


def define(heavydb):
    # Defines simple methods just to verify if NRT is working and if any there
    # are any symbol
    #
    # List of methods:
    # [X] List.append
    # [X] List.extend
    # [X] List.insert
    # [X] List.remove
    # [X] List.pop
    # [X] List.clear
    # [X] List.index
    # [X] List.count
    # [X] List.sort
    # [X] List.reverse
    # [X] List.copy

    @heavydb("i32(i32)", devices=['cpu'])
    def list_append(t):
        lst = list()
        lst.append('abc')
        return len(lst)

    @heavydb("i32(i32)", devices=['cpu'])
    def list_extend(t):
        lst = list()
        lst.extend('abc')
        return len(lst)

    @heavydb("i32(i32)", devices=['cpu'])
    def list_insert(t):
        lst = list()
        lst.insert(0, t)
        return len(lst)

    @heavydb("i32(i32)", devices=['cpu'])
    def list_remove(t):
        lst = list()
        lst.extend([t+1, t, t+2])
        lst.remove(t)
        return t in lst

    @heavydb("i32(i32)", devices=['cpu'])
    def list_pop(t):
        lst = list()
        lst.extend([t, t, t])
        lst.pop()
        return len(lst)

    @heavydb("i32(i32)", devices=['cpu'])
    def list_clear(t):
        lst = list()
        lst.extend([t, t, t])
        lst.clear()
        return len(lst) == 0

    @heavydb("i32(i32)", devices=['cpu'])
    def list_index(t):
        lst = list()
        lst.extend([t-1, t, t+1])
        return lst.index(t)

    @heavydb("i32(i32)", devices=['cpu'])
    def list_count(t):
        lst = list()
        lst.extend([t-1, t, t, t+1])
        return lst.count(t)

    @heavydb("bool(i32)", devices=['cpu'])
    def list_sort(t):
        lst = list()
        lst.extend([t+1, t, t-1])
        lst.sort()
        return lst[0] < lst[1] and lst[1] < lst[2]

    @heavydb("bool(i32)", devices=['cpu'])
    def list_reverse(t):
        lst = list()
        lst.extend([t+1, t, t-1])
        lst.reverse()
        return lst[0] < lst[1] and lst[1] < lst[2]

    @heavydb("i32(i32)", devices=['cpu'])
    def list_copy(t):
        lst = list()
        lst.extend([t-1, t, t+1])
        l2 = lst.copy()
        return len(l2)


@pytest.mark.parametrize('method,ans', [('append', 1), ('extend', 3), ('insert', 1),
                                        ('remove', 0), ('pop', 2), ('clear', 1),
                                        ('index', 1), ('count', 2), ('sort', 1),
                                        ('reverse', 1), ('copy', 3)])
def test_list_methods(heavydb, method, ans):
    _, result = heavydb.sql_execute(f"select list_{method}(1);")
    assert list(result) == [(ans,)]


@pytest.mark.parametrize('strategy', ['array', 'arraynull'])
def test_to_list(heavydb, strategy):

    @heavydb('int64[](int64[])', devices=['cpu'])
    def to_list(a):
        if a.is_null():
            other = Array(0, 'int64')
            other.set_null()
            return other
        else:
            lst = a.to_list()
            return Array(lst)

    heavydb.register()
    table = heavydb.table_name + strategy

    _, result = heavydb.sql_execute(f'select i8, to_list(i8) from {table}')
    result = list(zip(*result))
    expected, got = result
    for e, g in zip(expected, got):
        if e == []:
            # It is not possible to distinguish an empty list created by a null
            # array from one created by an empty array
            #   list(null array) -> []
            #   list(empty array) -> []
            assert g is None
        else:
            assert e == g



import pytest
from rbc.tests import heavydb_fixture
from rbc.errors import HeavyDBServerError

rbc_heavydb = pytest.importorskip("rbc.heavydb")
available_version, reason = rbc_heavydb.is_available()


@pytest.fixture(scope="module")
def heavydb():
    for o in heavydb_fixture(
        globals(), debug=False, load_test_data=False, minimal_version=(6, 2)
    ):
        try:
            o.sql_execute('select * from table(generate_series(0, 10, 1));')
        except Exception as e:
            msg = "Undefined function call 'generate_series' in SQL statement"
            if msg in str(e):
                pytest.skip('test requires server to be compiled with SYSTEM_TFS '
                            f'enabled ({e}):')
            else:
                raise
        define(o)
        yield o


def define(heavydb):
    @heavydb(
        "int32_t(TableFunctionManager, T start, T stop, K step, OutputColumn<T> generate_series)",
        T=["Timestamp"],
        K=["YearMonthTimeInterval", "DayTimeInterval"],
        devices=['cpu']
    )
    def rbc_generate_series(mgr, start, stop, step, series_output):
        if step.timeval == 0:
            return mgr.error_message("Timestamp division by zero")

        num_rows = step.numStepsBetween(start, stop) + 1
        if num_rows <= 0:
            mgr.set_output_row_size(0)
            return 0

        mgr.set_output_row_size(num_rows)

        for idx in range(num_rows):
            series_output[idx] = start + (step * idx)
        return num_rows


inputs = [
    (
        # 3-arg version, test with step of 1 second
        # Test non-named and named arg versions
        "TIMESTAMP(9) '1970-01-01 00:00:03.000000000'",
        "TIMESTAMP(9) '1970-01-01 00:00:07.000000000'",
        "INTERVAL '1' second",
        "ASC",
    ),
    (
        # Negative step and stop > start should return 0 rows
        "TIMESTAMP(9) '1970-01-01 00:00:02.000000000'",
        "TIMESTAMP(9) '1970-01-01 00:00:05.000000000'",
        "INTERVAL '-1' second",
        "ASC",
    ),
    (
        # Negative step and stop > start should return 0 rows
        "TIMESTAMP(9) '1970-01-01 00:00:05.000000000'",
        "TIMESTAMP(9) '1970-01-01 00:00:02.000000000'",
        "INTERVAL '1' second",
        "ASC",
    ),
    (
        # Negative step and stop == start should return 1 row (start)
        "TIMESTAMP(9) '1970-01-01 00:00:02.000000000'",
        "TIMESTAMP(9) '1970-01-01 00:00:02.000000000'",
        "INTERVAL '-1' second",
        "ASC",
    ),
    (
        # Positive step and stop == start should return 1 row (start)
        "TIMESTAMP(9) '1970-01-01 00:00:02.000000000'",
        "TIMESTAMP(9) '1970-01-01 00:00:02.000000000'",
        "INTERVAL '1' second",
        "ASC",
    ),
    (
        # 3-arg version - test step of 2 minutes
        # Test non-namned and named arg versions
        "TIMESTAMP(9) '1970-01-01 00:01:00.000000000'",
        "TIMESTAMP(9) '1970-01-01 00:10:00.000000000'",
        "INTERVAL '2' minute",
        "ASC",
    ),
    (
        # Series should be inclusive of stop value
        "TIMESTAMP(9) '2000-04-04 01:00:00.000000000'",
        "TIMESTAMP(9) '2000-04-04 09:00:00.000000000'",
        "INTERVAL '2' hour",
        "ASC",
    ),
    (
        # Negative hour step
        "TIMESTAMP(9) '1999-09-04 09:00:00.000000000'",
        "TIMESTAMP(9) '1999-09-04 01:00:00.000000000'",
        "INTERVAL '-1' hour",
        "DESC",
    ),
    (
        # 3-arg version - test month time intervals
        # Test non-namned and named arg versions
        "TIMESTAMP(9) '1970-01-01 00:00:00.000000000'",
        "TIMESTAMP(9) '1970-09-01 00:00:00.000000000'",
        "INTERVAL '1' month",
        "ASC",
    ),
    (
        # Negative month step
        "TIMESTAMP(9) '1999-09-04 09:00:00.000000000'",
        "TIMESTAMP(9) '1999-01-04 09:00:00.000000000'",
        "INTERVAL '-1' MONTH",
        "DESC",
    ),
    (
        # 3-arg version - test year time intervals
        # Test non-namned and named arg versions
        "TIMESTAMP(9) '1970-01-01 00:00:00.000000000'",
        "TIMESTAMP(9) '1979-01-01 00:00:00.000000000'",
        "INTERVAL '2' year",
        "ASC",
    ),
    (
        # Negative year step
        "TIMESTAMP(9) '1999-09-04 09:00:00.000000000'",
        "TIMESTAMP(9) '1991-09-04 09:00:00.000000000'",
        "INTERVAL '-1' YEAR",
        "DESC",
    ),
]


@pytest.mark.parametrize("start, stop, step, order", inputs)
def test_generate_time_series(heavydb, start, stop, step, order):

    if heavydb.version[:2] < (6, 2):
        pytest.skip('Requires HeavyDB version 6.2 or newer')

    RBC_FUNC = "rbc_generate_series"
    HEAVYDB_FUNC = "generate_series"

    non_named_arg_query = (
        "SELECT generate_series FROM TABLE({0}("
        f"{start},"
        f"{stop},"
        f"{step}))"
        f"ORDER BY generate_series {order};"
    )

    named_arg_query = (
        "SELECT generate_series FROM TABLE({0}("
        f"series_start=>{start},"
        f"series_stop=>{stop},"
        f"series_step=>{step}))"
        f"ORDER BY generate_series {order};"
    )

    for query in (non_named_arg_query, named_arg_query):
        _, rbc_result = heavydb.sql_execute(non_named_arg_query.format(RBC_FUNC))
        _, heavy_result = heavydb.sql_execute(non_named_arg_query.format(HEAVYDB_FUNC))
        assert list(rbc_result) == list(heavy_result)


invalid_inputs = [
    (
        # Step of 0 days is not permitted
        "TIMESTAMP(9) '1970-01-01 00:00:00.000000010'",
        "TIMESTAMP(9) '1970-01-01 00:00:00.000000020'",
        "INTERVAL '0' day",
        "Timestamp division by zero",
    ),
    (
        # Step of 0 months is not permitted
        "TIMESTAMP(9) '1970-01-01 00:00:00.000000010'",
        "TIMESTAMP(9) '1970-01-01 00:00:00.000000020'",
        "INTERVAL '0' month",
        "Timestamp division by zero",
    ),
]


@pytest.mark.parametrize("start, stop, step, error_msg", invalid_inputs)
def test_generate_series_invalid_inputs(heavydb, start, stop, step, error_msg):

    if heavydb.version[:2] < (6, 2):
        pytest.skip('Requires HeavyDB version 6.2 or newer')

    query = (
        "SELECT generate_series FROM TABLE(rbc_generate_series("
        f"{start},"
        f"{stop},"
        f"{step}));"
    )
    with pytest.raises(HeavyDBServerError) as exc:
        _, result = heavydb.sql_execute(query)

    assert exc.match(error_msg)



import pytest
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(6, 1), load_test_data=False):
        yield o


def test_get_registered_udtfs(heavydb):
    assert len(heavydb.table_function_names()) > 0


def test_get_runtime_udtfs_empty(heavydb):
    heavydb.unregister()
    assert len(heavydb.table_function_names(runtime_only=True)) == 0


def test_get_runtime_udtfs_non_empty(heavydb):
    heavydb.unregister()

    @heavydb('int32(Column<int>, OutputColumn<int>)')
    def foo(x, y):
        return 0

    heavydb.register()
    udtfs = heavydb.table_function_names(runtime_only=True)
    assert len(udtfs) == 1
    assert udtfs[0] == 'foo'


def test_get_udtf_details(heavydb):
    heavydb.unregister()

    @heavydb('int32(Column<int>, OutputColumn<int>)', devices=['cpu'])
    def my_udtf(x, y):
        return 0

    heavydb.register()
    details = heavydb.table_function_details('my_udtf')
    assert len(details) == 1

    @heavydb('int32(Column<int>, ConstantParameter, OutputColumn<int>)', devices=['cpu'])
    def another_udtf(x, m, y):
        return 0

    heavydb.register()
    details = heavydb.table_function_details('my_udtf', 'another_udtf')
    assert len(details) == 2


def test_get_invalid_udtf_details(heavydb):
    heavydb.reset()

    details = heavydb.table_function_details('invalid_udtf1234')
    assert len(details) == 0


def test_get_udfs_registered(heavydb):
    assert len(heavydb.function_names()) > 0


def test_get_runtime_udfs_empty(heavydb):
    heavydb.reset()
    assert len(heavydb.function_names(runtime_only=True)) == 0


def test_get_udf_details(heavydb):
    heavydb.reset()

    @heavydb('int32(int32)', devices=['cpu'])
    def incr(a):
        return a + 1

    heavydb.register()

    assert len(heavydb.function_details('incr')) == 1


def get_invalid_udf_details(heavydb):
    heavydb.reset()

    assert len(heavydb.function_details('invalid_udf_12345')) == 0



"""
Array API specification for set functions.

https://data-apis.org/array-api/latest/API_specification/set_functions.html
"""
from rbc.stdlib import Expose

__all__ = ["unique_all", "unique_counts", "unique_inverse", "unique_values"]

expose = Expose(globals(), "set_functions")


@expose.not_implemented("unique_all")
def _array_api_unique_all(x):
    """
    Returns the unique elements of an input array `x`, the first
    ocurring indices for each unique element in `x`, the indices from
    the set of unique elements that reconstruct `x`, and the corresponding
    counts for each unique element in `x`.
    """
    pass


@expose.not_implemented("unique_counts")
def _array_api_unique_counts(x):
    """
    Returns the unique elements of an input array `x` and the corresponding
    counts for each unique element in `x`.
    """
    pass


@expose.not_implemented("unique_inverse")
def _array_api_unique_inverse(x):
    """
    Returns the unique elements of an input array `x` and the indices
    from the set of unique elements that reconstruct `x`.
    """
    pass


@expose.not_implemented("unique_values")
def _array_api_unique_values(x):
    """
    Returns the unique elements of an input array `x`.
    """
    pass



import pytest
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), load_test_data=False):
        yield o


def test_TextEncodingNone_return(heavydb):
    heavydb.reset()

    with pytest.warns(SyntaxWarning):
        @heavydb('TextEncodingNone(int32, int32)')
        def make_abc(first, n):
            r = TextEncodingNone(n)  # noqa: F821
            for i in range(n):
                r[i] = first + i
            return r
    heavydb.reset()



import math
from rbc.externals import gen_codegen, dispatch_codegen
from numba.core.typing.templates import infer_global
from numba.core.imputils import lower_builtin
from numba.core.typing.templates import ConcreteTemplate, signature
from numba.types import float32, float64, int32, int64, uint64, intp


# Adding missing cases in Numba
@infer_global(math.log2)  # noqa: E302
class Math_unary(ConcreteTemplate):
    cases = [
        signature(float64, int64),
        signature(float64, uint64),
        signature(float32, float32),
        signature(float64, float64),
    ]


@infer_global(math.remainder)
class Math_remainder(ConcreteTemplate):
    cases = [
        signature(float32, float32, float32),
        signature(float64, float64, float64),
    ]


@infer_global(math.floor)
@infer_global(math.trunc)
@infer_global(math.ceil)
class Math_converter(ConcreteTemplate):
    cases = [
        signature(intp, intp),
        signature(int64, int64),
        signature(uint64, uint64),
        signature(float32, float32),
        signature(float64, float64),
    ]


booleans = []
booleans += [("isnand", "isnanf", math.isnan)]
booleans += [("isinfd", "isinff", math.isinf)]
booleans += [("isfinited", "finitef", math.isfinite)]

unarys = []
unarys += [("ceil", "ceilf", math.ceil)]
unarys += [("floor", "floorf", math.floor)]
unarys += [("fabs", "fabsf", math.fabs)]
unarys += [("exp", "expf", math.exp)]
unarys += [("expm1", "expm1f", math.expm1)]
unarys += [("erf", "erff", math.erf)]
unarys += [("erfc", "erfcf", math.erfc)]
unarys += [("tgamma", "tgammaf", math.gamma)]
unarys += [("lgamma", "lgammaf", math.lgamma)]
unarys += [("sqrt", "sqrtf", math.sqrt)]
unarys += [("log", "logf", math.log)]
unarys += [("log2", "log2f", math.log2)]
unarys += [("log10", "log10f", math.log10)]
unarys += [("log1p", "log1pf", math.log1p)]
unarys += [("acosh", "acoshf", math.acosh)]
unarys += [("acos", "acosf", math.acos)]
unarys += [("cos", "cosf", math.cos)]
unarys += [("cosh", "coshf", math.cosh)]
unarys += [("asinh", "asinhf", math.asinh)]
unarys += [("asin", "asinf", math.asin)]
unarys += [("sin", "sinf", math.sin)]
unarys += [("sinh", "sinhf", math.sinh)]
unarys += [("atan", "atanf", math.atan)]
unarys += [("atanh", "atanhf", math.atanh)]
unarys += [("tan", "tanf", math.tan)]
unarys += [("tanh", "tanhf", math.tanh)]
unarys += [("trunc", "truncf", math.trunc)]

binarys = []
binarys += [("copysign", "copysignf", math.copysign)]
binarys += [("atan2", "atan2f", math.atan2)]
binarys += [("pow", "powf", math.pow)]
binarys += [("fmod", "fmodf", math.fmod)]
binarys += [("hypot", "hypotf", math.hypot)]
binarys += [("remainder", "remainderf", math.remainder)]


def impl_unary(fname, key, typ):
    cpu = gen_codegen(fname)
    gpu = gen_codegen(f"__nv_{fname}")
    lower_builtin(key, typ)(dispatch_codegen(cpu, gpu))


def impl_binary(fname, key, typ):
    cpu = gen_codegen(fname)
    gpu = gen_codegen(f"__nv_{fname}")
    lower_builtin(key, typ, typ)(dispatch_codegen(cpu, gpu))


for fname64, fname32, key in unarys:
    impl_unary(fname64, key, float64)
    impl_unary(fname32, key, float32)


for fname64, fname32, key in binarys:
    impl_binary(fname64, key, float64)
    impl_binary(fname32, key, float32)


# manual mapping
def impl_ldexp():
    ldexp_cpu = gen_codegen('ldexp')
    ldexp_gpu = gen_codegen('__nv_ldexp')

    ldexpf_cpu = gen_codegen('ldexpf')
    ldexpf_gpu = gen_codegen('__nv_ldexpf')

    lower_builtin(math.ldexp, float64, int32)(dispatch_codegen(ldexp_cpu, ldexp_gpu))
    lower_builtin(math.ldexp, float32, int32)(dispatch_codegen(ldexpf_cpu, ldexpf_gpu))


impl_ldexp()

# CPU only:
# math.gcd
# math.degrees
# math.radians

# Missing:
# 'comb'
# 'dist'
# 'factorial'
# 'fsum'
# 'gcd'
# 'isclose'
# 'isqrt'
# 'frexp'
# 'modf'
# 'perm'
# 'prod'



"""Implement Heavydb Column type support

Heavydb Column type is the type of input/output column arguments in
UDTFs.
"""

__all__ = ['OutputColumn', 'Column', 'HeavyDBOutputColumnType', 'HeavyDBColumnType',
           'HeavyDBCursorType']

from llvmlite import ir
from rbc import typesystem
from .buffer import Buffer, HeavyDBBufferType, BufferType, BufferPointer
from .column_list import HeavyDBColumnListType
from . import text_encoding_none
from rbc.targetinfo import TargetInfo
from numba.core import extending
from numba.core import types as nb_types
from typing import Union


int8_t = ir.IntType(8)
int32_t = ir.IntType(32)


class HeavyDBColumnType(HeavyDBBufferType):
    """Heavydb Column type for RBC typesystem.
    """

    def postprocess_type(self):
        # if self.tostring().startswith('HeavyDBColumnType<HeavyDBArrayType'):
        #     from .column_array import HeavyDBColumnArrayType
        #     return self.copy(cls=HeavyDBColumnArrayType)
        # elif self.tostring().startswith('HeavyDBOutputColumnType<HeavyDBArrayType'):
        #     from .column_array import HeavyDBOutputColumnArrayType
        #     return self.copy(cls=HeavyDBOutputColumnArrayType)

        # re-wire the implementation of Column<T> to a subtype of Column
        from rbc import heavydb  # is there a better way to do this?
        geo_columns = ['GeoPoint', 'GeoMultiPoint',
                       'GeoLineString', 'GeoMultiLineString',
                       'GeoPolygon', 'GeoMultiPolygon']
        redirect_column_types = geo_columns + ['HeavyDBArray']
        kind = ['', 'Output']
        for p in kind:
            for s in redirect_column_types:
                column_name = f'HeavyDB{p}ColumnType<{s}'
                if self.tostring().startswith(column_name):
                    typename = f'HeavyDB{p}Column{s}Type'
                    # if self is a Column<Array<T>>, then, typename will be
                    # replaced to:
                    #  HeavyDB*ColumnHeavyDBArrayType -> HeavyDBColumnArrayType
                    typename = typename.replace('HeavyDBArray', 'Array')
                    cls = getattr(heavydb, typename)
                    return self.copy(cls=cls)

        return self

    def match(self, other):
        if type(self) is type(other):
            return self[0] == other[0]

    @property
    def buffer_extra_members(self):
        heavydb_version = TargetInfo().software[1][:3]
        if heavydb_version >= (6, 2) and self.element_type.tostring() == 'TextEncodingDict':
            return ('i8* string_dict_proxy_',)
        return ()


class HeavyDBOutputColumnType(HeavyDBColumnType):
    """Heavydb OutputColumn type for RBC typesystem.

    OutputColumn is the same as Column but introduced to distinguish
    the input and output arguments of UDTFs.
    """


class Column(Buffer):
    """
    RBC ``Column<T>`` type that corresponds to HeavyDB COLUMN

    In HeavyDB, a Column of type ``T`` is represented as follows:

    .. code-block:: c

        {
            T* ptr;
            int64_t sz;
            void* string_dict_proxy;  // available only if T == TextEncodingDict
        }

    """
    def dtype(self):
        """
        Data type of the array elements.
        """

    def getStringId(self, s: Union[str, 'text_encoding_none.TextEncodingNone']) -> int:  # noqa: E501
        """
        Return the string ID for the given string ``s``.

        .. note::
            Only available on ``CPU`` and ``Column<TextEncodingDict>``
        """

    def getString(self, id: int) -> int:
        """
        Return the string for the given ``id``.

        .. note::
            Only available on ``CPU`` and ``Column<TextEncodingDict>``
        """

    def __len__(self) -> int:
        """
        """


class OutputColumn(Column):
    pass


@extending.intrinsic
def heavydb_column_set_null_(typingctx, col_var, row_idx):
    # Float values are serialized as integers by HeavyDB
    # For reference, here is the conversion table for float and double
    #   FLOAT:  1.1754944e-38            -> 8388608
    #   DOUBLE: 2.2250738585072014e-308  -> 4503599627370496
    #                    ^                          ^
    #                 fp value                  serialized
    T = col_var.eltype
    sig = nb_types.void(col_var, row_idx)

    target_info = TargetInfo()
    null_value = target_info.null_values[str(T)]

    def codegen(context, builder, signature, args):
        data, index = args
        ptr = builder.extract_value(data, [0])

        ty = ptr.type.pointee
        nv = ir.Constant(ir.IntType(T.bitwidth), null_value)
        if isinstance(T, nb_types.Float):
            nv = builder.bitcast(nv, ty)
        builder.store(nv, builder.gep(ptr, [index]))

    return sig, codegen


@extending.overload_method(BufferType, 'set_null')
def heavydb_column_set_null(col_var, index):
    def impl(col_var, index):
        return heavydb_column_set_null_(col_var, index)
    return impl


@extending.intrinsic
def heavydb_column_is_null_(typingctx, col_var, row_idx):
    T = col_var.eltype
    sig = nb_types.boolean(col_var, row_idx)

    target_info = TargetInfo()
    null_value = target_info.null_values[str(T)]
    nv = ir.Constant(ir.IntType(T.bitwidth), null_value)

    def codegen(context, builder, signature, args):
        data, index = args
        ptr = builder.extract_value(data, [0])
        res = builder.load(builder.gep(ptr, [index]))

        if isinstance(T, nb_types.Float):
            res = builder.bitcast(res, nv.type)

        return builder.icmp_signed('==', res, nv)

    return sig, codegen


@extending.overload_method(BufferType, 'is_null')
def heavydb_column_is_null(col_var, row_idx):
    def impl(col_var, row_idx):
        return heavydb_column_is_null_(col_var, row_idx)
    return impl


@extending.intrinsic
def get_dict_proxy(typingctx, col_var):
    from .string_dict_proxy import StringDictionaryProxyNumbaType
    sig = StringDictionaryProxyNumbaType()(col_var)

    def codegen(context, builder, sig, args):
        [col] = args
        if col.type.is_pointer:
            col = builder.load(col)
        proxy = builder.extract_value(col, [2])
        return proxy

    return sig, codegen


@extending.overload_attribute(BufferType, "string_dict_proxy")
@extending.overload_attribute(BufferPointer, "string_dict_proxy")
def heavydb_column_string_dict_proxy(col_var):
    def impl(col_var):
        return get_dict_proxy(col_var)
    return impl


class HeavyDBCursorType(typesystem.Type):

    @classmethod
    def preprocess_args(cls, args):
        assert len(args) == 1
        params = []
        for p in args[0]:
            if not isinstance(p, (HeavyDBColumnType, HeavyDBColumnListType)):
                # map Cursor<T ...> to Cursor<Column<T> ...>
                c = p.copy()
                p = HeavyDBColumnType((c,), **c._params)
                c._params.clear()
            params.append(p)
        return (tuple(params),)

    @property
    def as_consumed_args(self):
        return self[0]



'''RBC TextEncodingDict type that corresponds to HeavyDB type TEXT ENCODED DICT.
'''

__all__ = ['HeavyDBTextEncodingDictType', 'TextEncodingDict']

from .metatype import HeavyDBMetaType
from rbc import typesystem
from numba.core import types as nb_types


class TextEncodingDictNumbaType(nb_types.Type):
    def __init__(self):
        super().__init__(name='TextEncodingDictNumbaType')


class HeavyDBTextEncodingDictType(typesystem.Type):
    """HeavyDB Text Encoding Dict type for RBC typesystem.
    """

    @property
    def __typesystem_type__(self):
        return typesystem.Type('int32')

    def postprocess_type(self):
        return self.params(shorttypename='TextEncodingDict')


class TextEncodingDict(object, metaclass=HeavyDBMetaType):
    """
    RBC ``TextEncodingDict`` type that corresponds to HeavyDB type TEXT ENCODED DICT.


    HeavyDB TextEncodingDict behaves like an int32_t.

    .. code-block:: c

        {
            int32_t value;
        }
    """
    pass



import pytest
from rbc.tests import heavydb_fixture


@pytest.fixture(scope="module")
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(5, 6), suffices=['']):
        define(o)
        yield o


def define(heavydb):
    @heavydb("int32(Column<int64>, RowMultiplier, OutputColumn<int64>)")
    def rbc_udtf_default_sizer_1(c, sizer, o):
        o[0] = 3
        return 1


def test_python_fn_with_default_sizer(heavydb):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5403")

    table = heavydb.table_name
    fn = "rbc_udtf_default_sizer_1"
    query = f"select * from table({fn}(cursor(select i8 from {table})));"
    _, result = heavydb.sql_execute(query)

    assert list(result) == [(3,)]


typs = (
    ("1a", "1", 1011),  # ct_udtf_default_sizer1__cpu_1
    ("1a", "1r", 1011),  # ct_udtf_default_sizer1__cpu_1
    ("1a", "1R", 1041),  # ct_udtf_default_sizer1__cpu_1
    ("1b", "11", 1013),  # ct_udtf_default_sizer1__cpu_3
    ("1b", "11r", 1013),  # ct_udtf_default_sizer1__cpu_3
    ("1b", "11R", 1046),  # ct_udtf_default_sizer1__cpu_3
    ("1c", "31a", 1342),  # ct_udtf_default_sizer1__cpu_8
    ("1c", "3r1a", 1342),  # ct_udtf_default_sizer1__cpu_8
    ("1c", "3R1a", 1372),  # ct_udtf_default_sizer1__cpu_8
    ("1d", "b1", 1241),  # ct_udtf_default_sizer1__cpu_10
    ("1d", "rb1", 1241),  # ct_udtf_default_sizer1__cpu_10
    ("1d", "Rb1", 1271),  # ct_udtf_default_sizer1__cpu_10
    ("2a", "1a", 3409),  # ct_udtf_default_sizer2__cpu_2
    ("2a", "1ar", 3409),  # ct_udtf_default_sizer2__cpu_2
    ("2a", "1aR", 3412),  # ct_udtf_default_sizer2__cpu_2
    ("2b", "11", 1003),  # ct_udtf_default_sizer2__cpu_4
    ("2b", "1r1", 1003),  # ct_udtf_default_sizer2__cpu_4
    ("2b", "1R1", 1006),  # ct_udtf_default_sizer2__cpu_4
    ("2c", "b1", 2552),  # ct_udtf_default_sizer2__cpu_11
    ("2c", "br1", 2552),  # ct_udtf_default_sizer2__cpu_11
    ("2c", "bR1", 2555),  # ct_udtf_default_sizer2__cpu_11
    ("3a", "1a", 1429),  # ct_udtf_default_sizer3__cpu_5
    ("3a", "1ra", 1429),  # ct_udtf_default_sizer3__cpu_5
    ("3a", "1Ra", 1729),  # ct_udtf_default_sizer3__cpu_5
    ("3b", "1a1", 1430),  # ct_udtf_default_sizer3__cpu_6
    ("3b", "1a1r", 1430),  # ct_udtf_default_sizer3__cpu_6
    ("3b", "1a1R", 1730),  # ct_udtf_default_sizer3__cpu_6
    ("4a", "11a", 1340),  # ct_udtf_default_sizer4__cpu_7
    ("4a", "1r1a", 1340),  # ct_udtf_default_sizer4__cpu_7
    ("4a", "1R1a", 1370),  # ct_udtf_default_sizer4__cpu_7
    ("4b", "1b", 1240),  # ct_udtf_default_sizer4__cpu_9
    ("4b", "r1b", 1240),  # ct_udtf_default_sizer4__cpu_9
    ("4b", "R1b", 1267),  # ct_udtf_default_sizer4__cpu_9
)


@pytest.mark.parametrize("suffix,kind,expected", typs)
def test_default_sizer(heavydb, suffix, kind, expected):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5403")

    codes = {
        "1": "i4",
        "2": "i4, i4",
        "3": "i4, i4, i4",
        "a": "cast(231 as INT)",
        "b": "132",
        "r": "1",  # sizer
        "R": "4",  # another sizer
    }

    table = heavydb.table_name
    fn = f"ct_udtf_default_sizer{suffix}"
    query = f"select * from table({fn}("
    for i, n in enumerate(kind):
        cols = codes[n]
        if n.isdigit():
            query += f"cursor(select {cols} from {table})"
        else:
            query += f"{cols}"
        query += ", " if i + 1 < len(kind) else ""
    query += "));"

    _, result = heavydb.sql_execute(query)
    result = list(result)

    assert result == [(expected,)], (result, query, kind)
